#cython: boundscheck=False, wraparound=False, initializedcheck=False

from finesse.cymath.complex cimport csqrt, cimag, COMPLEX_0
from finesse.cymath.math cimport acos, sqrt, PI
from finesse.cymath.math cimport sgn

from finesse.symbols import Symbol
from finesse.exceptions import BeamTraceException


cdef class CavityWorkspace(ElementWorkspace):
    def __init__(self, owner, BaseSimulation sim):
        cdef:
            TraceTree tree
            Py_ssize_t tree_idx

        super().__init__(sim, owner)

        self.tree = None
        for tree_idx in range(sim.trace_forest.size()):
            tree = sim.trace_forest.forest[tree_idx]

            if tree.dependency == owner and tree.is_source:
                if self.tree is not None:
                    raise BeamTraceException(
                        "Found multiple trees in the simulation "
                        f"trace forest corresponding to the same cavity {owner.name}"
                    )

                self.tree = tree

        self.is_changing = self.tree is not None

        self.abcd_x = owner.ABCDx
        self.abcd_y = owner.ABCDy

        ch_sym_length = owner._optical_length.expand_symbols().eval(keep_changing_symbols=True)
        self.length = float(ch_sym_length)
        if isinstance(ch_sym_length, Symbol):
            self.sym_length = cy_expr_new()
            cy_expr_init(self.sym_length, ch_sym_length)
        else:
            self.sym_length = NULL

        ch_sym_loss = owner._loss.expand_symbols().eval(keep_changing_symbols=True)
        self.loss = float(ch_sym_loss)
        if isinstance(ch_sym_loss, Symbol):
            self.sym_loss = cy_expr_new()
            cy_expr_init(self.sym_loss, ch_sym_loss)
        else:
            self.sym_loss = NULL

        ch_sym_finesse = owner._finesse.expand_symbols().eval(keep_changing_symbols=True)
        self.finesse = float(ch_sym_finesse)
        if isinstance(ch_sym_finesse, Symbol):
            self.sym_finesse = cy_expr_new()
            cy_expr_init(self.sym_finesse, ch_sym_finesse)
        else:
            self.sym_finesse = NULL

        ch_sym_fsr = owner._FSR.expand_symbols().eval(keep_changing_symbols=True)
        self.fsr = float(ch_sym_fsr)
        if isinstance(ch_sym_fsr, Symbol):
            self.sym_fsr = cy_expr_new()
            cy_expr_init(self.sym_fsr, ch_sym_fsr)
        else:
            self.sym_fsr = NULL

        ch_sym_fwhm = owner._FWHM.expand_symbols().eval(keep_changing_symbols=True)
        self.fwhm = float(ch_sym_fwhm)
        if isinstance(ch_sym_fwhm, Symbol):
            self.sym_fwhm = cy_expr_new()
            cy_expr_init(self.sym_fwhm, ch_sym_fwhm)
        else:
            self.sym_fwhm = NULL

        ch_sym_tau = owner._tau.expand_symbols().eval(keep_changing_symbols=True)
        self.tau = float(ch_sym_tau)
        if isinstance(ch_sym_tau, Symbol):
            self.sym_tau = cy_expr_new()
            cy_expr_init(self.sym_tau, ch_sym_tau)
        else:
            self.sym_tau = NULL

        ch_sym_pole = owner._pole.expand_symbols().eval(keep_changing_symbols=True)
        self.pole = float(ch_sym_tau)
        if isinstance(ch_sym_pole, Symbol):
            self.sym_pole = cy_expr_new()
            cy_expr_init(self.sym_pole, ch_sym_pole)
        else:
            self.sym_pole = NULL

        self.__update_geometric()

    def __dealloc__(self):
        cy_expr_free(self.sym_length)
        cy_expr_free(self.sym_loss)
        cy_expr_free(self.sym_finesse)
        cy_expr_free(self.sym_fsr)
        cy_expr_free(self.sym_fwhm)
        cy_expr_free(self.sym_tau)
        cy_expr_free(self.sym_pole)

    cdef void __update_geometric(self) noexcept:
        cdef:
            bint qx_ok, qy_ok

        qx_ok = self.update_eigenmode(cavity_plane.X)
        qy_ok = self.update_eigenmode(cavity_plane.Y)

        self.update_stability(cavity_plane.X)
        self.update_stability(cavity_plane.Y)
        self.is_stable_x = self.gx > 0.0 and self.gx < 1.0
        self.is_stable_y = self.gy > 0.0 and self.gy < 1.0
        self.is_stable = self.is_stable_x and self.is_stable_y

        if qx_ok or self.gx == 0 or self.gx == 1:
            self.update_rt_gouy(cavity_plane.X)
            self.update_Df(cavity_plane.X)
            self.update_S(cavity_plane.X)
        else:
            self.qx = COMPLEX_0
            self.rt_gouy_x = 0.0
            self.Df_x = 0.0
            self.Sx = 0.0

        if qy_ok or self.gy == 0 or self.gy == 1:
            self.update_rt_gouy(cavity_plane.Y)
            self.update_Df(cavity_plane.Y)
            self.update_S(cavity_plane.Y)
        else:
            self.qy = COMPLEX_0
            self.rt_gouy_y = 0.0
            self.Df_y = 0.0
            self.Sy = 0.0

    cdef void update(self) noexcept:
        # Update all the changing symbolic parameters
        if self.sym_length != NULL:
            self.length = cy_expr_eval(self.sym_length)
        if self.sym_loss != NULL:
            self.loss = cy_expr_eval(self.sym_loss)
        if self.sym_finesse != NULL:
            self.finesse = cy_expr_eval(self.sym_finesse)
        if self.sym_fsr != NULL:
            self.fsr = cy_expr_eval(self.sym_fsr)
        if self.sym_fwhm != NULL:
            self.fwhm = cy_expr_eval(self.sym_fwhm)
        if self.sym_tau != NULL:
            self.tau = cy_expr_eval(self.sym_tau)
        if self.sym_pole != NULL:
            self.pole = cy_expr_eval(self.sym_pole)

        # Update round-trip ABCD and dependent params if necessary
        if self.is_changing:
            self.tree.compute_rt_abcd(self.abcd_x, self.abcd_y)
            self.__update_geometric()

    cdef bint update_eigenmode(self, cavity_plane plane) noexcept:
        cdef:
            double[:, ::1] abcd = self.abcd_x if plane == cavity_plane.X else self.abcd_y
            double D_minus_A = abcd[1][1] - abcd[0][0]
            double minus_B = -abcd[0][1]
            double C = abcd[1][0]

            double half_inv_C
            complex_t sqrt_term
            complex_t lower, upper

            complex_t q

        if C == 0.0:
            return False

        half_inv_C = 0.5 / C

        sqrt_term = csqrt(D_minus_A * D_minus_A - 4 * C * minus_B)
        lower = (-D_minus_A - sqrt_term) * half_inv_C
        upper = (-D_minus_A + sqrt_term) * half_inv_C

        if cimag(lower) > 0:
            q = lower
        elif cimag(upper) > 0:
            q = upper
        else:
            return False

        if plane == cavity_plane.X:
            self.qx = q
        else:
            self.qy = q

        return True

    cdef void update_stability(self, cavity_plane plane) noexcept:
        cdef:
            double[:, ::1] abcd = self.abcd_x if plane == cavity_plane.X else self.abcd_y
            double A = abcd[0][0]
            double D = abcd[1][1]

            double g = 0.5 * (1 + 0.5 * (A + D))

        if plane == cavity_plane.X:
            self.gx = g
        else:
            self.gy = g

    cdef void update_rt_gouy(self, cavity_plane plane) noexcept:
        cdef:
            double[:, ::1] abcd = self.abcd_x if plane == cavity_plane.X else self.abcd_y
            double g = self.gx if plane == cavity_plane.X else self.gy
            double B = abcd[0][1]

            double gouy = 2.0 * acos(sgn(B) * sqrt(g))

        if plane == cavity_plane.X:
            self.rt_gouy_x = gouy
        else:
            self.rt_gouy_y = gouy

    cdef void update_Df(self, cavity_plane plane) noexcept:
        cdef:
            double gouy = self.rt_gouy_x if plane == cavity_plane.X else self.rt_gouy_y

            double df = 0.5 * self.fsr * gouy / PI

        if gouy > PI:
            df = self.fsr - df

        if plane == cavity_plane.X:
            self.Df_x = df
        else:
            self.Df_y = df

    cdef void update_S(self, cavity_plane plane) noexcept:
        cdef:
            double gouy = self.rt_gouy_x if plane == cavity_plane.X else self.rt_gouy_y

            double s = 0.5 * self.finesse * gouy / PI

        if gouy > PI:
            s = self.finesse - s

        if plane == cavity_plane.X:
            self.Sx = s
        else:
            self.Sy = s



from finesse.element_workspace cimport ElementWorkspace
from finesse.simulations.simulation cimport BaseSimulation
from finesse.utilities.collections cimport OrderedSet

ctypedef int (*fptr_c_gouy)(ABCDWorkspace) except -1


cdef class GouyFuncWrapper:
    cdef fptr_c_gouy func
    @staticmethod
    cdef GouyFuncWrapper make_from_ptr(fptr_c_gouy f)


cdef class ABCDWorkspace(ElementWorkspace):
    cdef readonly:
        GouyFuncWrapper fn_gouy_c  # C set Gouy phase function
        object fn_gouy_py  # Python set gouy function

    cpdef flag_changing_beam_parameters(self, OrderedSet changing_edges)
    cpdef update_map_data(self)



import pytest


@pytest.fixture
def fake_detector_cls(fake_element_cls):
    class FakeDetector(fake_element_cls):
        def __init__(self, name, target):
            super().__init__(name)
            self.target = target

    return FakeDetector


@pytest.fixture
def spec(spec, fake_element_adapter_factory, fake_element_cls):
    spec.register_element(fake_element_adapter_factory(fake_element_cls))
    spec.register_element(
        fake_element_adapter_factory(fake_detector_cls, "fake_detector")
    )
    return spec


def test_port(unbuilder, model, element_dump, fake_element_cls, fake_detector_cls):
    model.add(fake_element_cls("myel"))
    mydet = fake_detector_cls("mydet", target=model.myel.p1)
    model.add(mydet)
    dump = next(iter(element_dump("fake_detector", fake_detector_cls, model)))
    script = unbuilder.unbuild(dump)
    assert script == "fake_detector mydet target=myel.p1"


def test_node(unbuilder, model, element_dump, fake_element_cls, fake_detector_cls):
    model.add(fake_element_cls("myel"))
    mydet = fake_detector_cls("mydet", target=model.myel.p1.o)
    model.add(mydet)
    dump = next(iter(element_dump("fake_detector", fake_detector_cls, model)))
    script = unbuilder.unbuild(dump)
    assert script == "fake_detector mydet target=myel.p1.o"



import pytest
import finesse
import numpy as np
import finesse.analysis.actions as fa


@pytest.fixture
def model_solution():
    model = finesse.Model()
    model.parse(
        """
    l l1 P=1
    l l2 P=1 f=10M
    m m1 L=0 T=0
    link(l1, m1, l2)

    gauss g1 l1.p1.o w0=0.05 z=0
    modes(maxtem=2)
    """
    )
    model.l2.tem(0, 0, 0)
    model.l2.tem(0, 2, 1)
    return model, model.run("dc_fields(name='DC')")


def test_solution_name(model_solution):
    assert (
        model_solution[1]["DC"] is model_solution[1]
    ), "Didn't return the solution itself"


def test_get_node_name(model_solution):
    model, solution = model_solution
    _ = solution[("l1.p1.i", "l1.p1.o"), :, :]
    with pytest.raises(KeyError):
        _ = solution["l1.p1."]
    with pytest.raises(TypeError):
        _ = solution(["l1.p1.i", "l1.p1.o"])
    with pytest.raises(TypeError):
        _ = solution([model.l1.p1.i, model.l1.p1.o])


def test_values(model_solution):
    model, solution = model_solution
    # all nodes in 0Hz not in 00 == 0
    np.testing.assert_allclose(solution[:, 0, 1:], 0)
    np.testing.assert_allclose(solution[("l1.p1.o", "m1.p1.i"), 0, 0], 1)
    np.testing.assert_allclose(solution[(model.l1.p1.o, model.m1.p1.i), 0, 0], 1)
    # all nodes in 10MHz not in 02 == 0
    index = np.array([_ == [0, 2] for _ in solution.homs])
    np.testing.assert_allclose(solution[:, 1, ~index], 0)
    np.testing.assert_allclose(solution[("l2.p1.o", "m1.p2.i"), 1, index], 1)
    np.testing.assert_allclose(solution[(model.l2.p1.o, model.m1.p2.i), 1, index], 1)


def test_shapes(model_solution):
    model, sol = model_solution
    n_nodes = len(sol.nodes)
    n_freqs = 2
    n_homs = 6
    assert sol.fields.shape == (n_nodes, n_freqs, n_homs)
    assert sol["l2.p1.i"].shape == (1, n_freqs, n_homs)
    assert sol["l2.p1.i", :, 0].shape == (1, n_freqs)
    assert sol[("l2.p1.i", "m1.p1.o")].shape == (2, n_freqs, n_homs)
    assert sol[:, :, sol.homs.index([0, 1])].shape == (n_nodes, n_freqs)
    assert sol[5].shape == (n_freqs, n_homs)
    assert sol[1:3].shape == (2, n_freqs, n_homs)
    assert sol[1:3, 0].shape == (2, n_homs)
    assert sol[np.array([1, 3, 5, 7])].shape == (4, n_freqs, n_homs)
    idx = np.array([node.split(".")[-1] == "o" for node in sol.nodes])
    assert sol[idx].shape == (np.count_nonzero(idx), n_freqs, n_homs)


def test_node_string_section_equivalent(model_solution):
    model, sol = model_solution
    assert np.all(sol[model.l2.p1.i] == sol["l2.p1.i"])
    assert np.all(sol[model.l2.p1.i, :, 0] == sol["l2.p1.i", :, 0])
    assert np.all(sol[(model.l2.p1.i, model.m1.p1.o)] == sol[("l2.p1.i", "m1.p1.o")])


def test_change_param(model_solution):
    model, _ = model_solution
    Pnew = 100
    act0 = fa.Change({"l1.P": Pnew})
    act1 = fa.DCFields(name="DC")
    with model.temporary_parameters():
        model.l1.P = Pnew
        sol1 = model.run(act1)
    with model.temporary_parameters():
        sol2 = model.run(fa.Series(act0, act1))
    np.testing.assert_allclose(sol1.fields, sol2.fields)



"""Parameter tests."""

import pytest
from finesse.components import Beamsplitter, Laser, Mirror
from finesse.frequency import Fsig


@pytest.mark.parametrize(
    "component_type,args,parameter_default_status",
    (
        # Components without default parameters.
        (Laser, ["l1"], {k: False for k in ("P", "phase", "f")}),
        (Mirror, ["m1"], {k: False for k in ("R", "T", "L", "phi")}),
        # Components with default parameters.
        (Fsig, ["signal", 1e6], {"f": True}),
    ),
)
def test_default_model_parameters(component_type, args, parameter_default_status):
    """Model parameters listed as the default in their owner should know."""
    component = component_type(*args)

    for name, status in parameter_default_status.items():
        param = getattr(component, name)
        assert param.is_default_for_owner == status


@pytest.mark.parametrize(
    "component_type,args,params",
    (
        (Laser, ["l1"], ("P", "phase", "f", "signals_only")),
        (Mirror, ["m1"], ("R", "T", "L", "phi")),
        (Beamsplitter, ["bs"], ("R", "T", "L", "phi")),
    ),
)
def test_specific_parameter_help(component_type, args, params):
    """Check that the docstring for parameters gets replaced."""
    component = component_type(*args)
    for param in params:
        attr = getattr(component, param)
        assert attr.__doc__ != type(attr).__doc__


def test_autogen_name_parameter_full_name():
    import finesse

    model = finesse.Model()
    model.parse("m m1")
    model.parse("m m2")
    model.add(finesse.components.Space(None, model.m1.p1, model.m2.p1))
    assert model.m1.p1.space.L.full_name == "m1_p1__m2_p1.L"



import pytest

import numpy as np

from finesse import Model
from finesse.detectors import PowerDetector
from finesse.analysis.actions import Xaxis
from finesse.components import Gauss


@pytest.fixture
def forward_model():
    model = Model()
    model.parse(
        """
        l l1 P=1
        s s1 l1.p1 FI.p1
        isolator FI S=0
        """
    )
    model.add(PowerDetector("FI_p1", model.FI.p1.i))
    model.add(PowerDetector("FI_p2", model.FI.p2.o))
    return model


@pytest.fixture
def backward_model():
    model = Model()
    model.parse(
        """
        l l1 P=1
        s s1 l1.p1 FI.p2
        isolator FI S=0
        """
    )
    model.add(PowerDetector("FI_p1", model.FI.p1.o))
    model.add(PowerDetector("FI_p2", model.FI.p2.i))
    return model


@pytest.mark.parametrize("homs", (True, False))
class TestIsolator:

    def handle_homs(self, model, homs):
        if homs:
            model.add(Gauss("g1", model.l1.p1.o, w0=1e-3, z=0))
            model.modes(maxtem=1)
        return model

    def test_forward_complete_transmission(self, forward_model, homs):
        forward_model = self.handle_homs(forward_model, homs)
        sol = forward_model.run(Xaxis(forward_model.l1.P, "lin", 1, 10, 10))
        for det in ["FI_p1", "FI_p2"]:
            np.testing.assert_allclose(sol[det], np.linspace(1, 10, 11))

    def test_backward_complete_transmission(self, backward_model, homs):
        backward_model = self.handle_homs(backward_model, homs)
        sol = backward_model.run(Xaxis(backward_model.l1.P, "lin", 1, 10, 10))
        for det in ["FI_p1", "FI_p2"]:
            np.testing.assert_allclose(sol[det], np.linspace(1, 10, 11))

    def test_forward_suppression(self, forward_model, homs):
        forward_model = self.handle_homs(forward_model, homs)
        # should not be suppressing in the forward direction
        sol = forward_model.run(Xaxis(forward_model.FI.S, "lin", 1, 20, 10))
        for det in ["FI_p1", "FI_p2"]:
            np.testing.assert_allclose(sol[det], np.ones(11))

    def test_backward_suppression(self, backward_model, homs):
        backward_model = self.handle_homs(backward_model, homs)
        sol = backward_model.run(Xaxis(backward_model.FI.S, "lin", 1, 20, 10))
        np.testing.assert_allclose(sol["FI_p2"], np.ones(11))
        np.testing.assert_allclose(
            sol["FI_p1"], np.power(10 ** (-np.linspace(1, 20, 11) / 20), 2)
        )



from libc.stdlib cimport free, calloc
from libc.string cimport memcpy
from cpython.ref cimport Py_XINCREF, Py_XDECREF

from finesse.parameter cimport (
    Parameter,
    ParameterState,
)
from finesse.symbols import Symbol

class ElementValues:
    """Standard Python object which is used to store an Elements Parameter values.
    This is used in the default case where no optimised C class is provided."""
    pass


cdef class BaseCValues:
    """Base class that elements should use for storing their parameter
    values. This contains a generic method for storing a pointers to the
    double parameter values, which is intialised using `setup()`.

    Each Parameter of an element should result in a double. This won't handle
    any other types of data.
    """
    def __cinit__(self):
        self.ptr = NULL
        self.N = 0

    cdef setup(self, tuple params, Py_ssize_t data_size, double** data_start) :
        """Allocates the memory needed to store double values for the requested parameters.

        Examples
        --------
        An array of pointers to the double variables where each parameter should be copied across
        to.

        ctypedef (double*, double*) ptr_tuple_2 # data type of double points

        cdef class OpticalBandpassValues(BaseCValues):
            def __init__(self):
                cdef ptr_tuple_2 ptr = (&self.fc, &self.bandwidth) # array of ptrs to store parameter values at
                cdef tuple params = ("fc", "bandwidth") # names that match up in order with pointers
                self.setup(params, sizeof(ptr), <double**>&ptr) # call setup

        Parameters
        ----------
        params : tuple
            Tuple of Parameter names (case-sensitive) that an element has and needs to be stored in C values
        data_size : unsigned long
            number of double* pointers (or number of parameters)
        data_start : double**
            Pointer to array of double pointers where each parameter value should be stored
        """
        if self.ptr != NULL:
            raise ValueError("Memory already allocated")
        if data_start == NULL:
            raise ValueError("data_start == NULL")

        self.params = params
        self.N = data_size//sizeof(double)

        if len(self.params) != self.N:
            raise ValueError("Tuple of parameters and length of double pointers does not match")

        self.ptr = <double**> calloc(self.N, sizeof(double*))

        if not self.ptr:
            raise MemoryError()

        memcpy(self.ptr, data_start, data_size)

    def __dealloc__(self):
        if self.ptr != NULL:
            free(self.ptr)
            self.ptr = NULL

    cdef get_param_ptr(self, unicode name, double**ptr) :
        """Get a C pointer to where an elements parameter values should be stored.

        Parameters
        ----------
        name : unicode
            Name of the parameter
        ptr : double*
            Pointer to where the parameter pointer should be stored
        """
        if self.ptr == NULL:
            raise ValueError("Value storage pointers not set")
        if ptr == NULL:
            raise ValueError("ptr is NULL")

        idx = self.params.index(name)
        if self.ptr[idx] == NULL:
            raise ValueError(f"Pointer to {name} double is NULL")

        ptr[0] = self.ptr[idx]


cdef class ElementWorkspace:
    """
    This is the most basic workspace for a model element. It keeps track of the owner element and
    its parameter values in the `self.values` object.
    """
    def __cinit__(self, *args, **kwargs):
        self.num_changing_parameters = 0
        self.chprm = NULL
        self.chprm_target = NULL
        self.chprm_expr = NULL
        self.owner_id = -1
        self.first = True

    def __init__(self, object sim, object owner, object values=None):
        cdef:
            Parameter p
            int i
        self.owner = owner
        self.sim = sim

        if values is None:
            self.values = ElementValues()
            self.type_c_values = None
            self.is_c_values = False
        else:
            if not isinstance(values, BaseCValues):
                raise Exception("Values object should be a derivative of BaseCValues")

            self.values = values
            self.type_c_values = type(values)
            self.is_c_values = True

            # Here we setup for fast parameter setting by storing
            # the pyobject pointers to Parameters and also a double
            # pointer to the value

            # Make sure that numeric parameters come before symbolic
            # parameters so that the latter get eval'd to the correct
            # value when calling cy_expr_eval
            numeric_params = []
            symbolic_params = []
            for p in owner.parameters:
                if p.is_changing:
                    self.num_changing_parameters += 1
                    if p.state == ParameterState.Symbolic:
                        symbolic_params.append(p)
                    else:
                        numeric_params.append(p)

            params = numeric_params + symbolic_params
            if self.num_changing_parameters > 0:
                if self.chprm != NULL or self.chprm_target != NULL or self.chprm_expr != NULL:
                    raise MemoryError()

                self.chprm = <PyObject**> calloc(self.num_changing_parameters, sizeof(PyObject*))
                if not self.chprm:
                    raise MemoryError()

                self.chprm_target = <double**> calloc(self.num_changing_parameters, sizeof(double*))
                if not self.chprm_target:
                    raise MemoryError()

                self.chprm_expr = <cy_expr**> calloc(self.num_changing_parameters, sizeof(cy_expr*))
                if not self.chprm_expr:
                    raise MemoryError()

                i = 0
                for p in params:
                    if p.is_changing:
                        if p.state == ParameterState.Symbolic:
                            self.chprm_expr[i] = cy_expr_new()

                        self.chprm[i] = <PyObject*>p
                        Py_XINCREF(self.chprm[i])
                        (<BaseCValues>self.values).get_param_ptr(p.name, &self.chprm_target[i])
                        i += 1
                        # Stop this changing parameter from changing the type of
                        # variable it is (See #476). Problem comes about when a
                        # non-symbolic changes to a symbolic after this code has
                        # run, so then things like cy_expr_new have not been run.
                        p.__disable_state_type_change = True

    cpdef int compile_cy_exprs(self) except -1:
        if not self.num_changing_parameters:
            return 0

        for i in range(self.num_changing_parameters):
            if (<Parameter>self.chprm[i]).state == ParameterState.Symbolic:
                try:
                    # occasionally expressions simplify and become a constant
                    # so need to check
                    expr = (<Parameter>self.chprm[i]).value.expand_symbols().eval(keep_changing_symbols=True)
                    if isinstance(expr, Symbol):
                        if cy_expr_init(
                            self.chprm_expr[i], expr
                        ) == -1:
                            return -1
                except:
                    raise RuntimeError(f"Issue compiling cython expression :: {(<Parameter>self.chprm[i]).full_name}={(<Parameter>self.chprm[i]).value}")
        return 0

    cpdef update_parameter_values(self) :
        """Updates the `self.values` container which holds the latest values
        of this elements parameters.
        """
        cdef unicode p

        if self.first or not self.is_c_values:
            # First go just fill everything.
            # This is fairly slow but robust for any python object...
            vals, _ = self.owner._eval_parameters()
            for p in vals:
                if vals[p] is not None:
                    setattr(self.values, p, vals[p])
                else:
                    setattr(self.values, p, 0)
            self.first = False

        elif self.num_changing_parameters > 0:
            # Here we do some lower level setting of parameters using pointers
            # to the workspace.value.parameter double variable to speed things up
            for i in range(self.num_changing_parameters):
                if self.chprm_target[i] == NULL or self.chprm[i] == NULL:
                    raise MemoryError()

                if (<Parameter>self.chprm[i]).state == ParameterState.Numeric:
                    self.chprm_target[i][0] = (<Parameter>self.chprm[i]).__cvalue
                elif (<Parameter>self.chprm[i]).state == ParameterState.NONE:
                    self.chprm_target[i][0] = 0 # maybe should be NaN?
                elif (<Parameter>self.chprm[i]).state == ParameterState.Symbolic:
                    self.chprm_target[i][0] = cy_expr_eval(self.chprm_expr[i])
                    # Make sure __cvalue of symbolic parameters get updated too so that
                    # anything that relies on address of this uses correct value
                    (<Parameter>self.chprm[i]).__cvalue = self.chprm_target[i][0]
                else:
                    raise ValueError("Parameter state was unexpected")

    def __dealloc__(self):
        errors = []
        if self.num_changing_parameters > 0:
            for i in range(self.num_changing_parameters):
                if (<Parameter>self.chprm[i]).state == ParameterState.Symbolic:
                    cy_expr_free(self.chprm_expr[i])

                if self.chprm[i] == NULL:
                    errors.append(i)
                else:
                    Py_XDECREF(self.chprm[i])

            if self.chprm != NULL:
                free(self.chprm)
                self.chprm = NULL
            if self.chprm_expr != NULL:
                free(self.chprm_expr)
                self.chprm_expr = NULL
            if self.chprm_target != NULL:
                free(self.chprm_target)
                self.chprm_target = NULL

            if len(errors) > 0:
                raise MemoryError(f"unexpected self.chprm indices {errors} were NULL")

    def __repr__(self):
        try:
            return "<'{}' @ {} ({})>".format(
                self.owner.name, hex(id(self)), self.__class__.__name__
            )
        except:
            return super().__repr__()



from finesse.tree import TreeNode
from finesse.tree cimport TreeNode

cdef class ParameterChangingTreeNode(TreeNode):
    cdef public tuple parameters_changing

cdef class BaseSolution(ParameterChangingTreeNode):
    cdef public:
        double time # time taken to generate this solution



import pytest
from finesse.script.spec import KATSPEC, KatSpec, make_element
from finesse.components import Mirror


@pytest.mark.parametrize("directive", KATSPEC.directives.keys())
def test_directives_in_reference(directive, DIRECTIVES_REFERENCE):
    assert directive in DIRECTIVES_REFERENCE


def test_reference_directives_present(DIRECTIVES_REFERENCE):
    assert DIRECTIVES_REFERENCE == set(KATSPEC.directives.keys())


def test_python_class_name_present_in_spec():
    class Spec(KatSpec):
        def _register_constructs(self):
            self.register_element(make_element(int, "integer"))

    spec = Spec()
    assert "int" in spec.directives
    assert "integer" in spec.directives


def test_get_element_class():
    assert KATSPEC.get_element_class("Mirror") is Mirror


def test_get_element_class_missing():
    with pytest.raises(ValueError):
        KATSPEC.get_element_class("FooMirror")



from cpython.ref cimport PyObject

cdef extern from "tinyexpr.h" nogil:
    ctypedef struct te_expr:
        pass

    ctypedef struct te_variable:
        const char* name
        const void* address
        int type
        void* context

    cdef te_expr* te_compile(
        const char* expression,
        const te_variable* variables,
        int var_count,
        int* error
    ) noexcept

    cdef double te_eval(const te_expr* n) noexcept

    cdef void te_free(te_expr* n) noexcept


from .parameter cimport Parameter

# Wrapper for tinyexpr's expression and variables
# -> an array of ptrs to these get stored in ElementWorkspace
#    for fast evaluation of changing symbolic parameters
cdef struct cy_expr:
    te_expr* expr
    te_variable* variables

    const char* expression

    PyObject* byte_op_str

cdef cy_expr* cy_expr_new() except NULL nogil

cdef int cy_expr_init(cy_expr* ex, object operation) except -1

cdef void cy_expr_free(cy_expr* ex) noexcept

cdef double cy_expr_eval(cy_expr* ex) noexcept nogil



#cython: boundscheck=False, wraparound=False, initializedcheck=False

"""The TraceForest data structure used for representing propagating beams in a model.

Details on each class, method and function in this sub-module are provided mostly for
developers. Users should refer to :ref:`tracing_manual` for details on beam tracing,
:meth:`.Model.beam_trace` for the main method through which beam traces can be performed
on a model and :mod:`.tracing.tools` for the various beam propagation tools which the
beam tracing library provides.
"""

from finesse.cymath.math cimport float_eq
from finesse.cymath.gaussbeam cimport is_abcd_changing
from finesse.tracing.tree import get_tracing_tree
from finesse.tracing.tree cimport is_surface_refl
from finesse.components.node import NodeDirection

from itertools import chain
import logging
import numbers

from finesse.exceptions import BeamTraceException, TotalReflectionError, NoCouplingError
from finesse.utilities import refractive_index
from finesse.utilities.collections cimport OrderedSet
from finesse.utilities.collections import OrderedSet

LOGGER = logging.getLogger(__name__)


cdef class tree_container:
    """A container of TraceTree objects.

    Consists of a wrapper around a list of trees and a set of all
    the OpticalNode instances covered by this tree list. All of
    these attributes are read-only in the sense that they can only
    be accessed via C code.
    """
    def __init__(self, list trees=None):
        self.trees = []
        self.size = 0
        self.node_coverage = OrderedSet()

        # Support for creating a tree_container from pre-existing list of trees
        if trees:
            self.trees = trees.copy()
            self.size = len(self.trees)
            for tree in self.trees:
                self.node_coverage.update(tree.get_all_nodes())

    @property
    def trees(self):
        return self.trees[:]

    cdef clear(self) :
        self.trees.clear()
        self.node_coverage.clear()
        self.size = 0

    cdef append(self, TraceTree tree) :
        if tree is None:
            return

        self.trees.append(tree)
        self.size += 1
        self.node_coverage.update(tree.get_all_nodes())

    cdef remove(self, TraceTree tree) :
        if tree is None:
            return

        self.trees.remove(tree)
        self.size -= 1
        self.node_coverage.difference_update(tree.get_all_nodes())

    cdef _update_after_sub_remove(self, TraceTree sub_tree) :
        """Method to be called after removing sub-trees from the forest (via
        remove_left, remove_right on forest trees). Ensures that node_coverage
        remains consistent with forest state."""
        if sub_tree is None:
            return

        self.node_coverage.difference_update(sub_tree.get_all_nodes())

    def __getitem__(self, index):
        return self.trees[index]

    def __len__(self):
        return self.size




cdef class TraceForest:
    """A container structure which stores multiple :class:`.TraceTree` instances.

    The :class:`.Model` stores a TraceForest object which then represents the current
    tracing state of the configuration. Each time a :meth:`.Model.beam_trace` is called,
    either directly or indirectly, the TraceForest of the Model will be used to perform
    the tracing via propagation of the beam parameters through each tree. This is also
    detailed in :ref:`tracing_manual`.

    Determination of the ordering and overall structure of the TraceForest happens through
    the "planting" of the forest. By calling :meth:`.TraceForest.plant`, the forest is cleared
    and re-planted according to the ordered list of trace dependencies passed to this method.
    This is a step which is performed automatically in :meth:`.Model.beam_trace`, where this
    re-planting process only occurs under the following condition:

     * a connector has been added or removed since the last call,
     * the type of beam tracing has been switched from symmetric to
       asymmetric or vice-verase,
     * or the tracing priority (i.e. ordered list of trace dependencies)
       has changed in any way.

    In the initialisation process of building a simulation, a specialised version of a TraceForest
    is constructed from the model TraceForest using the ``TraceForest.make_changing_forest`` method.
    This inspects the model forest and selects only those trees, and branches of trees, which will
    have changing beam parameters during the simulation; i.e. due to some :class:`.GeometricParameter`
    being scanned. This new, "changing TraceForest" is then the optimised structure via which
    simulation-time beam traces (on changing beam parameter paths) are performed. More details on
    this, including additional powerful features that this changing forest provides, can be found
    in :ref:`tracing_manual`.

    .. rubric:: Special method support

    This class implements the following special methods: ``__getitem__``, ``__len__``, ``__iter__``,
    ``__next__`` and ``__contains__``, providing the following behaviour (assuming `forest` is an
    instance of this class):

      * ``tree = forest[x]`` - either get the :class:`.TraceTree` at index ``x`` (i.e. the x-th tree to be
        traced when performing a beam trace on the forest), OR if ``x`` is a :class:`.TraceDependency` get
        a list of all the trees in the forest which are associated with that dependency.
      * ``N_trees = len(forest)`` - the number of trees in the forest, equivalent to :meth:`.TraceForest.size`.
      * ``for tree in forest:`` - iteration support over the forest, in order of tracing priority.
      * ``flag = x in forest`` - check whether some object ``x`` is in the forest. This can be a :class:`.TraceTree`,
        an :class:`.OpticalNode`, a :class:`.Space` or a :class:`.Connector`. Equivalent to :meth:`.TraceForest.contains`.
    """
    def __init__(self, object model, bint symmetric, list trees=None):
        self.model = model
        self.symmetric = symmetric

        self.forest = tree_container(trees)
        self.dependencies = []

    def __deepcopy__(self, memo):
        raise RuntimeError("TraceForest instances cannot be copied.")

    ### Planting and clearing ###

    def plant(self, list trace_order):
        """Constructs and stores all the trace trees according to
        the order of dependencies in `trace_order`.

        Parameters
        ----------
        trace_order : list
            List of the dependency objects by priority of tracing.
        """
        from finesse.components import Cavity, Gauss

        self.clear()
        self.dependencies = trace_order.copy()

        # Always make the internal cavity trees first, these
        # will be generated in the order in which each cavity
        # appears in the trace_order dependency list
        cavities = list(filter(lambda x: isinstance(x, Cavity), self.dependencies))
        self._add_internal_cavity_trees(cavities)
        LOGGER.debug("TraceForest with internal Cavity trees: %s", self)

        # Handle ordering of internal cavity trees which overlap
        self._handle_overlapping_cavities()
        LOGGER.debug("TraceForest after overlapping cavities handled: %s", self)

        cdef OrderedSet internal_cav_nodes = self.forest.node_coverage.copy()
        cdef list gauss_nodes = []
        # Iterate through all dependencies by tracing order
        # and generate their associated trees
        for dep in self.dependencies:
            if isinstance(dep, Cavity):
                self._add_external_cavity_tree(dep)
            elif isinstance(dep, Gauss):
                mnode = next((n for n in gauss_nodes if n.space is dep.node.space), None)
                if mnode is not None:
                    mgauss = self.model.gausses[mnode]
                    # Cannot allow mismatches across spaces so explicitly ban
                    # separate Gauss objects at opposite ends of a space
                    if mnode.port is not dep.node.port:
                        raise BeamTraceException(
                            f"Gauss object {dep.name} is at the opposite end of a space to a "
                            f"previously defined Gauss ({mgauss.name}). Mode mismatches must occur "
                            "at connectors, not spaces, so disable or remove one of "
                            "these Gausses to propagate the intended beam.",
                        )
                    # Also prevent Gauss objects at opposite node of a port
                    # where a Gauss has already been defined
                    else:
                        raise BeamTraceException(
                            f"Gauss object {dep.name} is at the opposite node of a port with a "
                            f"previously defined Gauss ({mgauss.name}). Disable, or remove, one of "
                            "these Gausses to propagate the intended beam.",
                        )
                if dep.node in internal_cav_nodes:
                    raise BeamTraceException(
                        f"Gauss object {dep.name} is at an internal Cavity node "
                        f"({dep.node.full_name}). Disable, or remove, either the Gauss "
                        "or the corresponding Cavity to propagate the intended beam.",
                    )

                self._add_gauss_tree(dep)
                gauss_nodes.append(dep.node)
            else:
                raise TypeError(f"Unrecognised trace dependency type: {type(dep)}")
        LOGGER.debug(
            "TraceForest with internal, external Cavity trees and Gauss trees: %s",
            self,
        )

        # Remove common sub-trees sequentially
        self.trim()
        LOGGER.debug("TraceForest after trimming: %s", self)

        # Add branch trees at beam splitters where the nodes weren't
        # reachable from a previous trace tree
        cdef int missing = self._add_beamsplitter_branch_trees()
        LOGGER.debug("TraceForest after branch trees planted: %s", self)

        # If tracing is asymmetric and there are nodes which we can't
        # reach because of this (i.e. no path leading back to these
        # nodes), then we need to rectify this
        if not self.symmetric and missing:
            self._add_backwards_nonsymm_trees()
            # It's possible that there are still some branch nodes from beam splitters
            # that we couldn't reach at this point in asymmetric traces (e.g. a typical PDH
            # setup with a pick-off BS and one cav command), so we need to repeat the above
            # all over again to catch these trees
            if self._add_beamsplitter_branch_trees():
                self._add_backwards_nonsymm_trees()
            LOGGER.debug("TraceForest after backwards asymmetric trees added: %s", self)

        cdef OrderedSet diff = self.find_untraversed_nodes()
        if len(diff):
            raise BeamTraceException(
                f"The following optical nodes were not traced to. Ensure that a cavity or gaussian beam parameter is set that can reach them:\n{diff}"
            )

    cpdef void clear(self) noexcept:
        """Clears the trace forest, removing all trace trees."""
        self.forest.clear()

    cdef void trim(self) noexcept:
        cdef:
            Py_ssize_t i, j
            TraceTree tree1, tree2

            OrderedSet tree1_nodes
            OrderedSet trees_to_remove = OrderedSet()

        for i in range(self.forest.size):
            tree1 = self.forest[i]
            tree1_nodes = tree1.get_all_nodes()

            for j in range(i + 1, self.forest.size):
                tree2 = self.forest[j]
                # Don't attempt to trim common nodes for internal cavity trees
                if not tree2.is_source or tree2.dep_type == DependencyType.GAUSS:
                    # If any parts of tree2 overlap with tree1 then trim off
                    # these branches of tree2 to ensure uniqueness of trees
                    tree2.trim_at_nodes(tree1_nodes, self.symmetric)

                    if not tree2.is_source:
                        # Tree has no branches left so can be removed entirely
                        if tree2.left is None and tree2.right is None:
                            trees_to_remove.add(tree2)

        for tree in trees_to_remove:
            self.forest.remove(tree)

    cdef int _add_internal_cavity_trees(self, list cavities) except -1:
        cdef:
            Py_ssize_t Ncavs = len(cavities)
            Py_ssize_t cav_idx
            object cav

        for cav_idx in range(Ncavs):
            cav = cavities[cav_idx]

            self.forest.append(TraceTree.from_cavity(cav))

        return 0

    cdef int _handle_overlapping_cavities(self) except -1:
        cdef:
            Py_ssize_t i, j
            TraceTree tree1, tree2

            OrderedSet tree1_nodes
            list overlaps = []

        # Gather all the overlapping cavity combinations
        for i in range(self.forest.size):
            tree1 = self.forest[i]
            tree1_nodes = tree1.get_all_nodes()
            for j in range(i + 1, self.forest.size):
                tree2 = self.forest[j]
                for node in tree1_nodes:
                    if tree2._contains_node(node) or tree2._contains_node(node.opposite):
                        overlaps.append((tree1, tree2))
                        break

        # If there are no overlaps then nothing to do here
        if not overlaps:
            return 0

        # Make a flattened list of the overlapping tree combinations
        merged = list(chain.from_iterable(overlaps))
        # and merge the column slices of these combinations to get
        # the correct ordering before the operations below
        merged = merged[::2] + merged[1::2]

        # From this, create a list of these unique trees with the order
        # reversed such that trees which were added first from the cavity
        # trace order will overwrite the branches of the later trees which
        # intersect with them. This then guarantees that the trace_order given
        # to TraceForest.plant will be preserved for overlapping cavities.
        new_inner_order = list(dict.fromkeys(reversed(merged)))

        # Remove the internal cavity trees which are present in this
        # overlapping trees container...
        self.forest = tree_container(
            [tree for tree in self.forest.trees if tree not in new_inner_order]
        )
        # ... and then add them back in the order as outlined above
        for tree in new_inner_order:
            self.forest.append(tree)

        return 0

    cdef int _add_external_cavity_tree(self, object cav) except -1:
        cdef:
            dict exit_nodes
            object source, target

        exit_nodes = cav.get_exit_nodes()
        for source, target in exit_nodes.items():
            self.forest.append(
                TraceTree.from_node(
                    target,
                    cav,
                    self.symmetric,
                    pre_node=source,
                    exclude=self.forest.node_coverage,
                )
            )

        return 0

    cdef int _add_gauss_tree(self, object gauss) except -1:
        cdef:
            object gauss_node = gauss.node

            Py_ssize_t tree_idx
            TraceTree tree, new_tree, found, fp, fpp
            TraceTree rm_tree, tmp_tree
            OrderedSet trees_to_remove = OrderedSet()

        # Find branches in trees already planted that contain the gauss node
        # or its opposite direction
        for tree_idx in range(self.forest.size):
            tree = self.forest[tree_idx]

            found = tree.find_tree_at_node(gauss_node, include_opposite=True)
            if found is not None: # found a tree at the gauss node
                fp = found.parent
                if fp is not None: # this tree has a parent
                    # the parent node is an output i.e. a space exists between
                    # parent and the gauss -> so we want to remove parent too
                    if not fp.node.is_input:
                        fpp = fp.parent
                        if fpp is None: # parent has no parent so just remove the whole tree
                            trees_to_remove.add(tree)
                        else: # parent has a parent
                            if fp == fpp.left:
                                tmp_tree = fpp.remove_left()
                                self.forest._update_after_sub_remove(tmp_tree)
                            else:
                                tmp_tree = fpp.remove_right()
                                self.forest._update_after_sub_remove(tmp_tree)

                            # if the parent of parent has no remaining connections
                            # just remove the whole tree
                            if not fpp.is_source:
                                if fpp.left is None and fpp.right is None and fpp.parent is None:
                                    trees_to_remove.add(tree)
                    else: # parent node is an input (no space between parent -> gauss)
                        if found == fp.left:
                            tmp_tree = fp.remove_left()
                            self.forest._update_after_sub_remove(tmp_tree)
                        else:
                            tmp_tree = fp.remove_right()
                            self.forest._update_after_sub_remove(tmp_tree)

                        if not fp.is_source:
                            if fp.left is None and fp.right is None and fp.parent is None:
                                trees_to_remove.add(tree)
                else: # found tree has no parent so just remove the whole tree
                    trees_to_remove.add(tree)

        for rm_tree in trees_to_remove:
            self.forest.remove(rm_tree)

        new_tree = TraceTree.from_node(
            gauss_node,
            gauss,
            self.symmetric,
            is_source=True,
            exclude=self.forest.node_coverage,
        )
        self.forest.append(new_tree)

        # If the opposite direction of gauss node is not in the forward propagated
        # gauss tree then we need to make a tree from gauss.opposite too
        if (
            self.symmetric or
            (
                new_tree is not None and
                new_tree.find_tree_at_node(gauss_node.opposite) is None
            )
        ):
            # Backwards tree is not from the gauss node itself now
            # so leave is_source as False otherwise there would be
            # two sources from the same gauss node leading to problems
            back_new_tree = TraceTree.from_node(
                gauss_node.opposite,
                gauss,
                self.symmetric,
                exclude=self.forest.node_coverage - OrderedSet([gauss_node]),
            )

            if (
                back_new_tree is not None and
                (back_new_tree.left is not None or back_new_tree.right is not None)
            ):
                self.forest.append(back_new_tree)

        return 0

    cdef int _add_beamsplitter_branch_trees(self) except -1:
        cdef:
            double node_nr, pre_node_nr
            OrderedSet diff = self.find_untraversed_nodes()
            dict branch_start_nodes = {}

        from finesse.components.general import InteractionType

        if not diff:
            return 0

        tree = get_tracing_tree(self.model.optical_network)

        for node in diff:
            pre = list(tree.predecessors(node.full_name))
            node_nr = refractive_index(node)
            for pre_node_name in pre:
                pre_node = self.node_from_name(pre_node_name)
                pre_node_nr = refractive_index(pre_node)
                # If predecessor node is also in the set of unreachable nodes OR
                # the interaction type from pre_node -> node is a reflection and
                # the refractive indices at these ports are not the same (total
                # internal reflection) then skip it
                if (
                    pre_node in diff
                    or (
                        node.component.interaction_type(pre_node, node) == InteractionType.REFLECTION
                        and not float_eq(node_nr, pre_node_nr)
                    )
                ):
                    continue

                # Otherwise we want to store the predecessor node and its associated
                # dependency in a dict to make trees from later
                branch_start_nodes[node] = pre_node, self.find_dependency_from_node(pre_node)
                break

        if len(branch_start_nodes) == 0:
            # Shouldn't ever have a case where a branched node does not have
            # a predecessor (when doing symmetric planting), so if this
            # happens a bug has been encountered
            if self.symmetric:
                raise BeamTraceException(
                    f"These nodes were not reachable by the tracer and their beam parameter could not be set: {tuple(n.full_name for n in diff)}. "
                    "Ensure that you have set a cavity or gaussian beam parameter that can reach them."
                )
            # But if we're not a symmetric forest, then this will occur in
            # most cases as (for anything but extremely simple files) there
            # will be nodes which can't be reached directly, so inform on
            # return that this is the case
            else:
                return 1

        for node, (pre_node, dependency) in branch_start_nodes.items():
            self.forest.append(
                TraceTree.from_node(
                    node,
                    dependency,
                    self.symmetric,
                    pre_node=pre_node,
                    exclude=self.forest.node_coverage,
                )
            )

        return self._add_beamsplitter_branch_trees()

    cdef int _add_backwards_nonsymm_trees(self) except -1:
        cdef:
            TraceTree tree
            # twisted tree root and branch
            TraceTree ttr, ttb

            object comp
            bint is_dependency_changing

        # Get all the nodes we can't reach due to the asymmetric trace
        unreachable_nodes = self.find_untraversed_nodes()
        for node in unreachable_nodes:
            tree = self.find_tree_from_node(node.opposite)
            if tree is None:
                continue

            is_dependency_changing = tree.dependency.is_changing
            # Here we begin the process of making a "twisted tree" where
            # the left sub-tree node is actually a pre-coupling of the
            # parent tree node, this will allow us to apply the inverse
            # ABCD law transformation to left sub-tree node during tracing
            ttr = TraceTree.initialise(
                tree.node, tree.dependency, &is_dependency_changing
            )
            ttb = TraceTree.initialise(
                node, tree.dependency, &is_dependency_changing
            )

            if ttb.node.is_input:
                comp = ttb.node.component
            else:
                comp = ttb.node.space

            ttr.left = ttb
            ttb.parent = ttr

            try:
                # Check that there is a coupling from unreachable node -> opposite
                comp.check_coupling(ttb.node, ttr.node)

                try:
                    # Just to re-iterate, here we get the ABCD matrix coupling from
                    # the left tree to the parent tree (opposite to usual) in order
                    # to use this matrix in the inverse ABCD law transformation
                    ttr.set_left_abcd_x_memory(
                        comp.ABCD, ttb.node, ttr.node, direction="x", copy=False, retboth=True
                    )
                    ttr.set_left_abcd_y_memory(
                        comp.ABCD, ttb.node, ttr.node, direction="y", copy=False, retboth=True
                    )
                except TotalReflectionError:
                    raise

                ttr.is_left_surf_refl = is_surface_refl(comp, ttb.node, ttr.node)

                ttb.is_x_changing |= (
                    ttr.sym_left_abcd_x is not None
                    and is_abcd_changing(ttr.sym_left_abcd_x)
                )
                ttb.is_y_changing |= (
                    ttr.sym_left_abcd_y is not None
                    and is_abcd_changing(ttr.sym_left_abcd_y)
                )

                # Now mark the twisted tree root as an inverse transformation tree
                ttr.do_inv_transform = True
            except NoCouplingError:
                # No coupling exists from ttb.node -> ttr.node (typically means no
                # reflection coupling) so mark the root as needing a -q* operation
                # instead now as there is no other way to set the node q otherwise
                ttr.do_nonsymm_reverse = True

            self.forest.append(ttr)

    cpdef OrderedSet find_untraversed_nodes(self) :
        """Finds all the optical nodes in the model which are not
        covered by the trace forest."""
        cdef:
            OrderedSet nodes_traversed = self.forest.node_coverage
            OrderedSet nominal_diff = OrderedSet(self.model.optical_nodes).difference(nodes_traversed)
            OrderedSet real_diff = OrderedSet()

        if not self.symmetric:
            return nominal_diff

        for node in nominal_diff:
            if node.opposite not in nodes_traversed:
                real_diff.add(node)

        return real_diff

    ### Searching and attributes ###

    cpdef Py_ssize_t size(self) noexcept:
        """The number of trees in the forest."""
        return self.forest.size

    cpdef bint empty(self) noexcept:
        """Whether the forest is empty (no trees) or not."""
        return not self.forest.size

    def __getitem__(self, x):
        from finesse.components.trace_dependency import TraceDependency

        if isinstance(x, numbers.Integral):
            return self.forest[x]
        if isinstance(x, TraceDependency):
            return self.trees_of_dependency(x)

        raise TypeError(
            "Expected sub-script of type: Integral or TraceDependency, but "
            f"received {x} of type {type(x)}"
        )

    def __len__(self):
        return self.forest.size

    def __iter__(self):
        return iter(self.forest.trees)

    def __next__(self):
        return next(self.forest.trees)

    def __contains__(self, o):
        return self.contains(o)

    cpdef bint contains(self, object o) noexcept:
        """Whether the forest contains the specified object, determined recursively for
        each tree within the forest.

        Parameters
        ----------
        o : [:class:`.TraceTree` | :class:`.OpticalNode` | :class:`.Space` | :class:`.Connector`]
            The object to search for in the forest.

        Returns
        -------
        flag : bool
            True if `o` is in the forest, False otherwise.
        """
        cdef:
            Py_ssize_t tree_idx
            TraceTree tree

        for tree_idx in range(self.forest.size):
            tree = self.forest[tree_idx]

            if tree.contains(o):
                return True

        return False

    cdef list _get_trees_upred(self, bint (*predicate)(TraceTree)):
        cdef:
            Py_ssize_t tree_idx
            TraceTree tree

            list trees = []

        for tree_idx in range(self.forest.size):
            tree = self.forest[tree_idx]

            if predicate(tree):
                trees.append(tree)

        return trees

    cdef list _get_trees_bpred(self, bint (*predicate)(TraceTree, object), object o):
        cdef:
            Py_ssize_t tree_idx
            TraceTree tree

            list trees = []

        for tree_idx in range(self.forest.size):
            tree = self.forest[tree_idx]

            if predicate(tree, o):
                trees.append(tree)

        return trees

    cpdef list trees_of_dependency(self, object dependency) :
        """Get a list of all the :class:`.TraceTree` instances with the
        associated trace `dependency` object.

        Parameters
        ----------
        dependency : :class:`.TraceDependency`
            A trace dependency object.

        Returns
        -------
        trees : list
            A list of all trace trees with dependency equal to above object.
        """
        return self._get_trees_bpred(dep_match_pred, dependency)

    cdef list get_internal_cavity_trees(self) :
        return self._get_trees_upred(internal_cav_pred)

    @property
    def internal_cavity_trees(self):
        return self.get_internal_cavity_trees()

    cdef list get_external_cavity_trees(self) :
        return self._get_trees_upred(external_cav_pred)

    @property
    def external_cavity_trees(self):
        return self.get_external_cavity_trees()

    cdef list get_gauss_trees(self) :
        return self._get_trees_upred(gauss_pred)

    @property
    def gauss_trees(self):
        return self.get_gauss_trees()

    cpdef TraceTree find_tree_from_node(self, object node) :
        """Given an optical node, this finds the :class:`.TraceTree` instance
        corresponding to this node (if one exists).

        Parameters
        ----------
        node : :class:`.OpticalNode`
            An optical node.

        Returns
        -------
        tree : :class:`.TraceTree`
            The tree corresponding to `node`, or ``None`` if none found.
        """
        cdef:
            Py_ssize_t tree_idx
            TraceTree tree, found

        for tree_idx in range(self.forest.size):
            tree = self.forest[tree_idx]

            found = tree.find_tree_at_node(node, self.symmetric)
            if found is not None:
                return found

        return None

    cpdef object find_dependency_from_node(self, object node, bint raise_not_found=True) :
        """Finds the dependency object associated with the optical `node`.

        If no tree is found associated with this node, and `raise_not_found` is true,
        then a ``BeamTraceException`` is raised. Otherwise `None` is returned.

        Parameters
        ----------
        node : :class:`.OpticalNode`
            An optical node.

        raise_not_found : bool, optional; default: True
            Raises a BeamTraceException if no dependency found. Returns `None` if False.
        """
        cdef:
            Py_ssize_t tree_idx
            TraceTree tree, found

        for tree_idx in range(self.forest.size, 0, -1):
            tree = self.forest[tree_idx - 1]

            found = tree.find_tree_at_node(node, self.symmetric)
            if found is not None:
                return found.dependency

        if raise_not_found:
            raise BeamTraceException(
                f"Could not find a dependency object associated with node {node.full_name} in the trace forest."
            )
        else:
            return None

    cdef object node_from_name(self, name) :
        return self.model.network.nodes[name]["weakref"]()

    ### Changing geometric parameter forest algorithms ###

    cpdef TraceForest make_changing_forest(self) :
        """Constructs a new TraceForest from this forest, consisting
        of only the trees which will have changing beam parameters.

        This method is called in BaseSimulation._initialise for setting
        up the simulation trace forest used for efficient beam tracing.
        """
        cdef:
            Py_ssize_t tree_idx
            TraceTree tree, chtree

            list changing_trees = []
            bint branch_added = False

        for tree_idx in range(self.forest.size):
            tree = self.forest[tree_idx]
            branch_added = False
            # For branched beamsplitter trees, need to see if the root
            # node opposite is already present in the changing trees
            # list --- if so then add the branch tree as this will also
            # be changing
            for chtree in changing_trees:
                if chtree.find_tree_at_node(tree.node.opposite) is not None:
                    changing_trees.append(tree)
                    branch_added = True
                    break

            if not branch_added:
                # From this tree, obtain the broadest sub-trees which
                # will have changing beam parameters
                changing_trees.extend(tree.get_broadest_changing_subtrees())

        cdef Py_ssize_t Nchanging_nominal = len(changing_trees)
        cdef OrderedSet roots = OrderedSet() # Parents of changing trees
        cdef OrderedSet trees_to_remove = OrderedSet()
        for tree_idx in range(Nchanging_nominal):
            tree = changing_trees[tree_idx]
            # Now set each changing tree to the parent tree so that
            # the root is used in beam tracing...
            if tree.parent is not None:
                # ... but only do this for parents not yet added
                # otherwise could get duplicate trees
                if tree.parent not in roots:
                    changing_trees[tree_idx] = tree.parent
                    roots.add(tree.parent)
                else:
                    # This tree will already be encapsulated by the previous
                    # parent so just mark it to be removed
                    trees_to_remove.add(tree)

        for tree in trees_to_remove:
            changing_trees.remove(tree)

        cdef TraceForest changing_forest = TraceForest(self.model, self.symmetric, changing_trees)

        return changing_forest

    ### Automatic mode mismatch coupling determination ###

    cpdef tuple find_potential_mismatch_couplings(self, TraceForest other=None) :
        """Retrieves the node couplings which are potentially mode mismatched. If
        `other` is not given then the couplings which are local to this forest only
        will be found, otherwise couplings between this forest and `other` will
        be retrieved.

        If this forest is asymmetric, then calling this method is equivalent to
        calling :meth:`.TraceForest.find_intersection_couplings`.

        This method is used internally for obtaining all the possible mode mismatch
        couplings between a changing trace forest (held by a modal simulation) and
        the main model trace forest.

        Parameters
        ----------
        other : :class:`.TraceForest`
            Find dependencies from a different trace forest than this one
            when checking for mode mismatch couplings.

        Returns
        -------
        couplings : tuple
            A tuple of the node couplings where each element is ``(from_node, to_node)``.
        """
        from finesse.components.general import InteractionType

        intersect_couplings = self.find_intersection_couplings(other)

        cdef list refls = []
        cdef list fake_refl_mismatches = []
        cdef tuple other_mrefls
        cdef object[:, ::1] refl_abcd_sym_x
        cdef object[:, ::1] refl_abcd_sym_y
        # If we're doing a symmetric trace then self-reflections from mirror-type components
        # can be potential mode mismatch couplings so need to add these too
        if self.symmetric:
            refls.extend(self.get_mirror_reflection_couplings())
            for node1, node2 in refls:
                opp_surface_onode = None
                # Find the output node on the other side of the surface
                for n1s_name in list(self.model.optical_network.successors(node1.full_name)):
                    n1s = self.node_from_name(n1s_name)
                    if node1.component.interaction_type(node1, n1s) == InteractionType.TRANSMISSION:
                        opp_surface_onode = n1s
                        break

                # If none found then nothing else needs to be done for this coupling
                if opp_surface_onode is None:
                    continue

                # Reflection coupling on other side of surface is in the potential
                # mismatch couplings so this one can remain too
                if (opp_surface_onode.opposite, opp_surface_onode) in refls:
                    continue

                # Get the dependencies associated with the two sides of the surface
                dep1 = self.find_dependency_from_node(node1)
                dep2 = self.find_dependency_from_node(opp_surface_onode, raise_not_found=False)
                if dep2 is None and other is not None:
                    dep2 = other.find_dependency_from_node(opp_surface_onode)

                # Finally, if the dependencies of the trees on both sides are the same
                # then this isn't really a mismatch coupling (as beam params on both sides
                # are guaranteed to be mode matched in such a case) so mark it to be removed
                if dep1 is dep2:
                    fake_refl_mismatches.append((node1, node2))

            for fnodes in fake_refl_mismatches:
                refls.remove(fnodes)

            # In addition, if the model trace forest is specified via other
            # then we need to check for reflection couplings here which
            # impinge against connectors with changing ABCDs as these will
            # also be potential mode mismatch couplings
            if other is not None:
                other_mrefls = other.get_mirror_reflection_couplings(
                    skip_dependencies=self.dependencies,
                )

                for node1, node2 in other_mrefls:
                    # Don't add the coupling if we already determined that it
                    # was a "fake" mismatch coupling (see above)
                    if (node1, node2) in fake_refl_mismatches:
                        continue

                    comp = node1.component
                    # Get the symbolic ABCDs upon reflection from the surface...
                    refl_abcd_sym_x = comp.ABCD(
                        node1, node2, "x", copy=False, symbolic=True
                    )
                    refl_abcd_sym_y = comp.ABCD(
                        node1, node2, "y", copy=False, symbolic=True
                    )

                    # ... and check if they're changing, if so we have another
                    # possible mode mismatch coupling which needs to be added
                    if is_abcd_changing(refl_abcd_sym_x) or is_abcd_changing(refl_abcd_sym_y):
                        refls.append((node1, node2))

        return intersect_couplings + tuple(OrderedSet(refls))


    cpdef tuple find_intersection_couplings(self, TraceForest other=None) :
        """Finds the node couplings at which trees with differing trace dependencies intersect.

        Parameters
        ----------
        other : :class:`.TraceForest`
            Find dependencies from a different trace forest than this one
            when checking for intersections.

        Returns
        -------
        couplings : tuple
            A tuple of the node couplings where each element is ``(from_node, to_node)``.
        """
        cdef:
            Py_ssize_t otree_idx
            TraceTree tree

            list couplings = []

        if not self.forest.size:
            return ()

        for otree_idx in range(self.forest.size):
            tree = self.forest[otree_idx]

            # An internal cavity trace can have intersection couplings
            # from any input node in the cavity path
            if tree.is_source and tree.dep_type == DependencyType.CAVITY:
                last_nodes = [n for n in tree.get_all_nodes() if n.is_input]
            else:
                # Get the final *input* nodes at the end of the tree, across all branches
                last_nodes = tree.get_last_input_nodes()
                # Need to also add the reverse node of the tree if it's a source
                # tree so that intersections are checked in the other propagation
                if tree.is_source and tree.node.opposite not in last_nodes:
                    last_nodes.append(tree.node.opposite)

            for node in last_nodes:
                # Obtain the successor nodes of this (if any) from the network
                succ_nodes = list(self.model.optical_network.successors(node.full_name))

                for snode_name in succ_nodes:
                    snode = self.node_from_name(snode_name)
                    # Then for each successor node find its TraceTree in the forest
                    # and obtain the dependency that this relies upon
                    if other is None:
                        snode_dep = self.find_dependency_from_node(snode)
                    else:
                        snode_dep = other.find_dependency_from_node(snode)

                    # If this dependency is not the same object as the original tree's
                    # dependency then we've found an intersection so add this coupling
                    if snode_dep is not tree.dependency:
                        # Here we do some sanity checks on the coupling we've found
                        node.component
                        snode.component
                        if node.component is not snode.component:
                            raise BeamTraceException(
                                "Found an intersection coupling "
                                f"{node.full_name} -> {snode.full_name} which "
                                "does not occur across the same connector. Mode "
                                "mismatches must not occur across Spaces in Finesse."
                            )
                        node.component.check_coupling(node, snode)

                        couplings.append((node, snode))
                        # Add the reverse coupling too if it exists
                        try:
                            node.component.check_coupling(snode.opposite, node.opposite)
                            couplings.append((snode.opposite, node.opposite))
                        except:
                            pass

        return tuple(OrderedSet(couplings))

    cpdef tuple get_mirror_reflection_couplings(
        self,
        bint ignore_internal_cavities=True,
        list skip_dependencies=None,
    ) :
        """Get the node couplings in the forest which correspond to self-reflections
        from mirror-like components.

        Parameters
        ----------
        ignore_internal_cavities : bool, default: True
            Ignore the node couplings inside cavities.

        skip_dependencies : list
            Optional list of trees to skip based on their dependencies.

        Returns
        -------
        couplings : tuple
            A sequence of tuples consisting of the node1 -> node2 self
            reflection couplings.
        """
        cdef:
            Py_ssize_t tree_idx
            TraceTree tree

            list couplings = []

        if skip_dependencies is None:
            skip_dependencies = []

        # Gather the nodes which are part of any cavity cycle
        cdef OrderedSet internal_cav_nodes = OrderedSet()
        for tree in self.get_internal_cavity_trees():
            internal_cav_nodes.update(tree.get_all_nodes())

        for tree_idx in range(self.forest.size):
            tree = self.forest[tree_idx]

            if tree.dependency in skip_dependencies:
                continue

            if ignore_internal_cavities:
                # The tree is an internal cavity tree itself so skip the whole thing
                if tree.is_source and tree.dep_type == DependencyType.CAVITY:
                    continue

                # If it's a tree coming directly from an internal cavity tree then
                # get the reflection couplings only from the second node onwards
                if tree.node in internal_cav_nodes:
                    if tree.left is not None:
                        couplings.extend(tree.left.get_mirror_reflection_couplings())
                else:
                    couplings.extend(tree.get_mirror_reflection_couplings())
            else:
                couplings.extend(tree.get_mirror_reflection_couplings())

        return tuple(OrderedSet(couplings))

    def get_nodes_with_changing_q(self):
        """For a given TraceForest this method will determine which optical nodes in a
        model will have a changing complex beam parameter.

        This relies on element Parameters having their ``is_tunable`` or ``is_changing``
        being set as ``True``. In such cases the Model will construct a simulation where
        the calculations dependent on these parameters will be recomputed at multiple
        steps.

        Returns
        -------
        q_changing_nodes : set{OpticalNode}
            Set of of OpticalNodes which will have a changing complex beam parameter
            during a simulation.
        """
        model = self.model
        q_changing_nodes = OrderedSet()
        # 1) Get changing geometric parameters
        changing_geometric = [p for el in model.elements.values() for p in el.parameters if p.is_geometric and p.is_changing]
        # 2) Get elements with changing geometric parameters
        changing_geometric_elements = {p.owner for p in changing_geometric}
        # 3) Can't tell yet which changing geometric parameter will affect which
        #    coupling, so grab all output nodes which might carry new q away
        output_nodes = [n for el in changing_geometric_elements for n in el.optical_nodes if n.direction == NodeDirection.OUTPUT]
        if len(output_nodes) == 0:
            return q_changing_nodes

        # 4) Are any nodes in a cavity trace? If so they will propgate to other intersecting tree
        affected_cavities = {tree.dependency for n in output_nodes for tree in model.trace_forest.internal_cavity_trees if tree.contains(n)}

        if len(affected_cavities) == 0:
            # Changing nodes are not part of a cavity, but can still have
            # downstream affects.
            # For each node, find the downstream tree to see which nodes it will effect
            output_trees = tuple(model.trace_forest.find_tree_from_node(n) for n in output_nodes)
            if len(output_trees) > 0:
                # Merge all the nodes into one unique set
                q_changing_nodes = OrderedSet.union(*(t.get_all_nodes() for t in output_trees))
        else:
            # otherwise get all the nodes in each of the cavities being affected
            q_changing_nodes = OrderedSet.union(*(tree.get_all_nodes() for tree in chain.from_iterable(model.trace_forest.trees_of_dependency(c) for c in affected_cavities)))

        if model.trace_forest.symmetric and len(q_changing_nodes) > 0:
            # If it's symmetric tracing we need to make
            # sure pairs of nodes are marked
            q_changing_nodes = OrderedSet.union(q_changing_nodes, {n.opposite for n in q_changing_nodes})

        return q_changing_nodes

    ### Drawing ###

    cpdef draw_by_dependency(self) :
        """Draws the forest as a string representation.

        All the trees in the forest are sorted by their dependency and
        stored in the resultant string by these dependency sub-headings. Each
        tree also has its index (i.e. tracing priority) stored in the string
        above the drawn tree.

        Returns
        -------
        forest_str : str
            A string representation of the forest, sorted by dependency with
            tracing priority indices displayed for each tree.
        """
        cdef:
            Py_ssize_t tree_idx
            TraceTree tree

            list dependencies = []

            all_trees_str = ""

        for tree_idx in range(self.forest.size):
            tree = self.forest[tree_idx]
            if tree.dependency not in dependencies:
                dependencies.append(tree.dependency)

        for dependency in dependencies:
            # Make sub-heading for each dependency, giving its name and type name
            all_trees_str += f"\nDependency: {dependency.name} [{type(dependency).__name__}]\n\n"
            for tree_idx in range(self.forest.size):
                tree = self.forest[tree_idx]
                if tree.dependency == dependency:
                    # Give the index of the tree in the forest before drawing each tree
                    # -> indicates tracing order of the tree
                    all_trees_str += (
                        f"  (Index: {tree_idx})\n" + tree.draw(left_pad="    ") + "\n\n"
                    )

        return all_trees_str

    cpdef draw(self) :
        """Draws the forest, by trace priority, as a string representation.

        The order in which trees appear in this string represents the order in which
        they will be traced during the beam tracing algorithm.

        In the rare cases where a subsequent tree contains a duplicate node (from an
        earlier tree), the latter tree trace will overwrite the former. This is only
        applicable to configurations with overlapping cavities, and this overwriting
        behaviour will take account of the desired cavity ordering given by the user.

        Returns
        -------
        forest_str : str
            A string representation of the ordered forest.
        """
        cdef:
            Py_ssize_t tree_idx
            TraceTree tree

            object last_dep = None
            all_trees_str = ""

        for tree_idx in range(self.forest.size):
            tree = self.forest[tree_idx]
            dep = tree.dependency

            if dep is not last_dep:
                all_trees_str += f"\nDependency: {dep.name} [{type(dep).__name__}]\n\n"
            else:
                all_trees_str += "\n"

            all_trees_str += tree.draw(left_pad="  ") + "\n"

            last_dep = dep

        return all_trees_str

    def __str__(self):
        return self.draw()

    ### Propagating beams ###

    cpdef dict trace_beam(self) :
        """Performs a "model-time" beam trace on all trace trees.

        This method is called internally by :meth:`.Model.beam_trace`. One should
        use that method to get a more complete representation of the tracing of
        the beam through a model.

        Returns
        -------
        trace : dict
            Dictionary of `node: (qx, qy)` mappings where `node` is each
            :class:`.OpticalNode` instance and `qx, qy` are the beam
            parameters in the tangential and sagittal planes, respectively,
            at these nodes.
        """
        cdef:
            Py_ssize_t tree_idx
            TraceTree tree

            double lambda0 = self.model.lambda0

            dict trace = {}

        for tree_idx in range(self.forest.size):
            tree = self.forest[tree_idx]
            if tree.is_source:
                trace.update(tree.trace_beam(lambda0, self.symmetric))
            else:
                tree.propagate(trace, lambda0, self.symmetric)

        return trace


### Useful low-level predicates for TraceTree filtering ###

cdef bint dep_match_pred(TraceTree tree, object dependency) noexcept:
    return tree.dependency is dependency

cdef bint internal_cav_pred(TraceTree tree) noexcept:
    return tree.is_source and tree.dep_type == DependencyType.CAVITY

cdef bint external_cav_pred(TraceTree tree) noexcept:
    return not tree.is_source and tree.dep_type == DependencyType.CAVITY

cdef bint gauss_pred(TraceTree tree) noexcept:
    return tree.is_source and tree.dep_type == DependencyType.GAUSS



"""
Testing unbuilding of parameters that are defined as minus a symbolic
reference to another parameter.
"""

from finesse import Model
from finesse.script.generator import KatUnbuilder


def test_unparse_negative_reference_kat():
    """Negative reference defined in katscript."""
    model = Model(
        """
        m m1 phi=22
        m m2 phi=-m1.phi
        """
    )

    unbuilder = KatUnbuilder()
    script = unbuilder.unbuild(model)
    assert "mirror m2 R=0.5 T=0.5 L=0.0 phi=-m1.phi" in script


def test_unparse_negative_reference_python():
    """Negative reference defined in python."""
    model = Model(
        """
        m m1 phi=22
        m m2
        """
    )
    model.m2.phi = -model.m1.phi.ref

    unbuilder = KatUnbuilder()
    script = unbuilder.unbuild(model, prefer_keywords=True)
    assert "mirror m2 R=0.5 T=0.5 L=0.0 phi=-m1.phi" in script



"""Configuration tools."""

# NOTE: this module gets imported by `finesse` directly, so cannot itself import from
# `finesse` and cannot import packages that themselves import from `finesse`.
import importlib.resources
import logging
import os
from configparser import RawConfigParser
from pathlib import Path

from . import datastore
from .utilities import option_list

_PACKAGE_LOGGER = logging.getLogger(__package__)

show_progress_bars = False


def config_instance():
    """The Finesse configuration object for the current session.

    Returns
    -------
    :class:`configparser.RawConfigParser`
        The Finesse configuration object.
    """
    return datastore.init_singleton(_FinesseConfig)


def configure(plotting=False, jupyter_tracebacks=None, progress_bars=False):
    """Configure Finesse runtime options.

    Parameters
    ----------
    plotting : bool, optional
        Initialise Finesse plot theme for display.

    jupyter_tracebacks : bool, optional
        Show full tracebacks in Finesse errors when using Finesse in IPython. This
        setting does not work reliably in other environments.

    progress_bars : bool, optional
        When True progress bars will be displayed when available.

    See Also
    --------
    :func:`finesse.plotting.tools.init`
    """
    # NOTE: Before modifying this function, note that it may be called multiple times
    # during the execution of Finesse and should therefore remain idempotent.

    if plotting:
        from .plotting import init as init_plotting

        init_plotting(mode="display")

    if jupyter_tracebacks is not None:
        from .env import show_tracebacks

        show_tracebacks(jupyter_tracebacks)

    global show_progress_bars
    show_progress_bars = progress_bars


def autoconfigure():
    """Automatically configure runtime options based on the environment."""
    from .env import is_interactive

    kwargs = {}

    if is_interactive():
        # The user has imported Finesse inside a notebook or similar interactive
        # session. Configure some aspects of Finesse automatically: pretty plots and
        # suppress tracebacks.
        kwargs["plotting"] = False  # FIXME: docs don't build when True!
        kwargs["jupyter_tracebacks"] = False

    configure(**kwargs)


class _FinesseConfig(RawConfigParser):
    """The built-in and user configuration for Finesse.

    Do not instantiate this class directly; use :func:`config_instance`.
    """

    # Order in which user configs are loaded.
    _USER_CONFIG_LOAD_ORDER = ["user_config_path", "cwd_config_path"]

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.write_user_config()
        self._load_finesse_configs()

    @classmethod
    def user_config_paths(cls):
        return {
            config: getattr(cls, config)() for config in cls._USER_CONFIG_LOAD_ORDER
        }

    @classmethod
    def user_config_path(cls):
        return cls.user_config_dir() / "usr.ini"

    @classmethod
    def cwd_config_path(cls):
        return Path.cwd() / "finesse.ini"

    @classmethod
    def user_config_dir(cls):
        r"""The path to the user's config directory for Finesse.

        The exact path is determined by the current platform and the presence of certain
        environment variables:

        .. rubric:: Windows

        A folder called ``finesse`` in the folder pointed to by the environment variable
        ``%APPDATA`` (usually ``%HOMEPATH%\AppData\Roaming``).

        .. rubric:: POSIX (including macOS and WSL)

        A directory called ``finesse`` inside either the path pointed to by the
        environment variable ``XDG_CONFIG_HOME`` or, if that value cannot be found or is
        empty, ``~/.config``.

        Returns
        -------
        :py:class:`pathlib.Path` or None
            The path to the Finesse config directory.

        Raises
        ------
        :py:class:`RuntimeError`
            If no config directory can be determined.
        """
        from .env import IS_WINDOWS

        if IS_WINDOWS:
            try:
                config_dir = Path(os.environ["APPDATA"])
            except KeyError:
                # NOTE: we assume %APPDATA% always exists on any normal Windows machine,
                # which should be the case. If it's not, we might need to change Finesse
                # to handle having no user config path.
                raise RuntimeError(
                    r"The %APPDATA% environment variable is required for Finesse to "
                    r"store user configuration, but it was not found. Please ensure "
                    r"this environment variable exists in the environment in which "
                    r"Finesse is being run."
                )
        else:
            # Path.home() raises RuntimeError if no home is found.
            config_dir = Path(
                os.environ.get("XDG_CONFIG_HOME", Path.home() / ".config")
            )

        return config_dir / "finesse"

    @classmethod
    def write_user_config(cls, force=False):
        """Copy the default config files to the user's config directory."""
        logger = logging.getLogger(__name__)
        user_config_path = cls.user_config_path()

        if force or not user_config_path.is_file():
            # Copy barebone user config file contents into user's user config path.
            logger.info(f"Writing user config file to {user_config_path}.")

            user_config_path.parent.mkdir(parents=True, exist_ok=True)
            with user_config_path.open("wb") as fobj:
                fobj.write(
                    importlib.resources.files(__package__)
                    .joinpath("usr.ini.dist")
                    .read_bytes()
                )

    def _load_finesse_configs(self):
        """Read the built-in and any user configuration files.

        The built-in configuration is loaded first, then the user configuration files
        are loaded in the order specified in :attr:`.USER_PATHS`. This means user
        configuration options can overwrite built-in options, and options from later
        paths in :attr:`.USER_PATHS` can overwrite options from earlier paths.
        """
        user_config_paths = self.user_config_paths().values()

        # Load the bundled configuration.
        self.read_string(
            (importlib.resources.files("finesse") / "finesse.ini").read_text(),
            source="<bundled finesse.ini>",
        )

        # Parse all configurations, from lowest to highest priority.
        parsed = self.read(user_config_paths)

        if not parsed:
            paths = option_list(user_config_paths)

            raise ConfigNotFoundError(f"Could not find user config files at {paths}.")


class ConfigNotFoundError(Exception):
    """Indicates a Finesse configuration could not be loaded."""



"""Legacy parser 1:n attribute mappings."""

import finesse


def test_g_attr():
    """Test that using `g` correctly maps to `user_gouy_x` and `user_gouy_y`."""
    model = finesse.Model()
    model.parse_legacy(
        """
        l l1 1 0 n0
        s s1 1 n0 n1
        attr s1 g 90
        """
    )

    assert model.s1.user_gouy_x == 90
    assert model.s1.user_gouy_y == 90


def test_gx_attr():
    """Test that using `gx` correctly maps to `user_gouy_x`."""
    model = finesse.Model()
    model.parse_legacy(
        """
        l l1 1 0 n0
        s s1 1 n0 n1
        attr s1 gx 90
        """
    )

    assert model.s1.user_gouy_x == 90


def test_gy_attr():
    """Test that using `gy` correctly maps to `user_gouy_y`."""
    model = finesse.Model()
    model.parse_legacy(
        """
        l l1 1 0 n0
        s s1 1 n0 n1
        attr s1 gy 90
        """
    )

    assert model.s1.user_gouy_y == 90



"""KatScript specification.

This defines supported KatScript syntax and maps it to Finesse Python classes via
:class:`.ItemAdapter` objects.
"""

from __future__ import annotations

import logging
from collections import ChainMap

from .. import components, detectors, frequency, locks, symbols
from ..analysis import actions, noise
from ..components import electronics, mechanical
from ..config import config_instance
from ..model import Model
from .adapter import (
    AnalysisDocumenter,
    AnalysisDumper,
    AnalysisFactory,
    AnalysisSetter,
    Argument,
    ArgumentDump,
    ArgumentType,
    CommandDump,
    CommandMethodDocumenter,
    CommandMethodSetter,
    CommandPropertyDocumenter,
    CommandPropertyDumper,
    CommandPropertySetter,
    ElementDocumenter,
    ElementDump,
    ElementDumper,
    ElementFactory,
    ElementSetter,
    ItemAdapter,
    ItemDumper,
    ItemSetter,
    SignatureArgumentMixin,
)

LOGGER = logging.getLogger(__name__)


def add_name_to_other_names(name: str, kwargs: dict) -> dict:
    """For 'make_element' and 'make_analysis', adds the name of the python class to the
    list of 'other_names', which get added in 'ItemAdapter' to the list of aliases.
    Allows 'Xaxis' to be recognized as 'xaxis' and 'Beamsplitter' as 'beamsplitter'.

    Parameters
    ----------
    cls_name : str
        Name of the class
    kwargs : dict
        kwargs meant to be passed to ItemAdapter

    Returns
    -------
    dict
        kwargs with 'other_names' modified to include the class name.
    """
    other_names = kwargs.setdefault("other_names", [])
    other_names.append(name)
    return kwargs


def make_element(element_type, full_name, build_last=False, ref_args=None, **kwargs):
    """Create a standard element adapter.

    Use this for elements that follow the normal behaviour: they are of type
    :class:`.ModelElement`, their constructor signature defines their attributes, and
    the attributes are available with the same name in the object.
    """
    kwargs = add_name_to_other_names(element_type.__name__, kwargs)
    return ItemAdapter(
        full_name,
        getter=ElementDumper(item_type=element_type, ref_args=ref_args),
        factory=ElementFactory(item_type=element_type, last=build_last),
        setter=ElementSetter(item_type=element_type),
        documenter=ElementDocumenter(item_type=element_type),
        **kwargs,
    )


def make_analysis(analysis_type, full_name, ref_args=None, **kwargs):
    """Create a standard analysis adapter.

    Use this for analyses that follow the normal behaviour: they are of type
    :class:`.Action`, their constructor signature defines their attributes, and the
    attributes are available with the same name in the object.
    """
    kwargs = add_name_to_other_names(analysis_type.__name__, kwargs)
    return ItemAdapter(
        full_name,
        getter=AnalysisDumper(item_type=analysis_type, ref_args=ref_args),
        factory=AnalysisFactory(item_type=analysis_type),
        setter=AnalysisSetter(item_type=analysis_type),
        documenter=AnalysisDocumenter(item_type=analysis_type),
        singular=True,
        **kwargs,
    )


class GaussDumper(ItemDumper):
    """Dumper for the :class:`.Gauss` component.

    Gauss components accept many different forms of argument to define the beam
    parameter but only store these internally as q parameters. This dumper therefore
    tries to figure out how the parameters were defined by the user (either via Python
    or KatScript) and generates parameters based on that.
    """

    def __init__(self):
        super().__init__(item_type=components.Gauss)

    def __call__(self, adapter, model):
        for gauss in model.get_elements_of_type(self.item_type):
            parameters = {
                "node": ArgumentDump("node", value=gauss.node, kind=ArgumentType.ANY),
                "priority": ArgumentDump(
                    "priorty", value=gauss.priority, default=0, kind=ArgumentType.ANY
                ),
            }

            # Map of available gauss call parameters to attributes.
            # FIXME: move these to Gauss itself.
            gauss_attr_map = {
                "w0": gauss.qx.w0,
                "w0x": gauss.qx.w0,
                "w0y": gauss.qy.w0,
                "z": gauss.qx.z,
                "zx": gauss.qx.z,
                "zy": gauss.qy.z,
                "zr": gauss.qx.zr,
                "zrx": gauss.qx.zr,
                "zry": gauss.qy.zr,
                "w": gauss.qx.w,
                "wx": gauss.qx.w,
                "wy": gauss.qy.w,
                "Rc": gauss.qx.Rc,
                "Rcx": gauss.qx.Rc,
                "Rcy": gauss.qy.Rc,
                "S": gauss.qx.S,
                "Sx": gauss.qx.S,
                "Sy": gauss.qy.S,
                "q": gauss.qx.q,
                "qx": gauss.qx.q,
                "qy": gauss.qy.q,
            }

            # Add additional parameters specified by the user.
            for param in gauss._specified_params:
                value = gauss_attr_map[param]
                parameters[param] = ArgumentDump(
                    param,
                    value=value,
                    default=None,
                    kind=ArgumentType.KEYWORD_ONLY,
                )

            yield ElementDump(
                element=gauss,
                adapter=adapter,
                parameters=parameters,
                is_default=all(param.is_default for param in parameters.values()),
            )


class VariableDumper(ItemDumper):
    """Variable for :class:`finesse.components.general.Variable`."""

    def __init__(self):
        super().__init__(item_type=components.Variable)

    def __call__(self, adapter, model):
        for variable in model.parameters:
            parameters = {
                "value": ArgumentDump(
                    "value", value=variable.value, kind=ArgumentType.POS_ONLY
                ),
                "units": ArgumentDump(
                    "units",
                    value=variable.units,
                    kind=ArgumentType.KEYWORD_ONLY,
                    default="",
                ),
                "description": ArgumentDump(
                    "description",
                    value=variable.description,
                    kind=ArgumentType.KEYWORD_ONLY,
                    default=None,
                ),
                # Not sure how to handle dtype
                # "dtype": ArgumentDump("dtype", value=variable.datatype, kind=ArgumentType.KEYWORD_ONLY),
                "is_geometric": ArgumentDump(
                    "is_geometric",
                    value=variable.is_geometric,
                    kind=ArgumentType.KEYWORD_ONLY,
                    default=False,
                ),
                "changeable_during_simulation": ArgumentDump(
                    "changeable_during_simulation",
                    value=variable.changeable_during_simulation,
                    kind=ArgumentType.KEYWORD_ONLY,
                    default=True,
                ),
            }

            yield ElementDump(
                element=variable,
                adapter=adapter,
                parameters=parameters,
                is_default=False,
            )


class FsigArgumentsMixin:
    def arguments(self, **kwargs):
        # FIXME: these would ideally come from Fsig's constructor, but we can't inspect
        # Cython types.
        return {"f": Argument("f", ArgumentType.ANY)}


class FsigDumper(ItemDumper):
    def __init__(self):
        super().__init__(item_type=frequency.Fsig)

    def __call__(self, adapter, model):
        # FIXME: these would ideally come from Fsig's constructor, but we can't inspect
        # Cython types.
        default = None
        is_default = model.fsig.f.value == default
        parameters = {"f": ArgumentDump("f", value=model.fsig.f, default=default)}

        yield CommandDump(adapter=adapter, parameters=parameters, is_default=is_default)


class FsigSetter(FsigArgumentsMixin, ItemSetter):
    def __init__(self):
        super().__init__(item_type=frequency.Fsig)

    def __call__(self, model, argskwargs):
        args, kwargs = argskwargs
        allargs = list(args) + list(kwargs.values())
        assert len(allargs) == 1
        model.fsig.f = allargs[0]


class FsigDocumenter(FsigArgumentsMixin, CommandMethodDocumenter):
    def __init__(self):
        super().__init__(item_type=frequency.Fsig)

    def argument_descriptions(self):
        descriptions = super().argument_descriptions()
        # The KatScript call to fsig doesn't use the name.
        descriptions.pop("name")

        return descriptions


class ModesDumper(SignatureArgumentMixin, ItemDumper):
    def __init__(self):
        super().__init__(item_type=Model.modes)

    def __call__(self, adapter, model):
        # Grab the values from :meth:`.Model.modes_setting` but grab the defaults from
        # the :meth:`.Model.modes` signature.
        sigparams = self.arguments()
        parameters = {
            key: ArgumentDump(
                key,
                value=value,
                default=sigparams[key].default,
                kind=sigparams[key].kind,
            )
            for key, value in model.modes_setting.items()
        }

        # Ensure the "modes" parameter exists so that, if "default valued" commands are
        # requested, a modes command with an "off" parameter is dumped.
        modes_other_defaults = [sigparams["modes"].default]
        if modes := parameters.get("modes"):
            # The `modes` parameter accepts "off" via KatScript as the same meaning as
            # the Python default (`None`).
            modes.default = "off"
            modes.other_defaults = modes_other_defaults
        else:
            # Ensure there is at least an "off".
            parameters["modes"] = ArgumentDump(
                "modes",
                value="off",
                default="off",
                other_defaults=modes_other_defaults,
                kind=sigparams["modes"].kind,
            )

        yield CommandDump(
            adapter=adapter,
            parameters=parameters,
            is_default=all(param.is_default for param in parameters.values()),
        )


class TEMDumper(SignatureArgumentMixin, ItemDumper):
    def __init__(self):
        super().__init__(item_type=components.Laser.tem)

    def __call__(self, adapter, model):
        """(args, kwargs) tuples for each defined TEM mode."""
        sigparams = self.arguments()

        for laser in model.get_elements_of_type(components.Laser):
            for (n, m), (factor, phase) in laser.power_coeffs.items():
                otherkv = {"n": n, "m": m, "factor": factor, "phase": phase}
                other_parameters = {
                    key: ArgumentDump(
                        key,
                        value=value,
                        default=sigparams[key].default,
                        kind=sigparams[key].kind,
                    )
                    for key, value in otherkv.items()
                }

                parameters = {
                    "laser": ArgumentDump("laser", value=laser),
                    **other_parameters,
                }

                current = {(n, m): (factor, phase)}
                yield CommandDump(
                    adapter=adapter,
                    parameters=parameters,
                    is_default=current.items() <= laser.DEFAULT_POWER_COEFFS.items(),
                )


class TEMSetter(CommandMethodSetter):
    def __init__(self):
        super().__init__(item_type=components.Laser.tem)

    def __call__(self, _, argskwargs):
        args, kwargs = argskwargs

        if "laser" in kwargs:
            laser = kwargs.pop("laser")
        else:
            laser, *args = args

        self.item_type(laser, *args, **kwargs)


class TEMDocumenter(CommandMethodDocumenter):
    def __init__(self):
        super().__init__(item_type=components.Laser.tem)

    ##
    # Override some inherited methods to add the laser argument, which is not part of
    # :meth:`.Laser.tem` (well, technically it's `self`, but we want to call it
    # "laser").

    def arguments(self):
        return {
            "laser": Argument(name="laser", kind=ArgumentType.ANY),
            **super().arguments(),
        }

    def argument_descriptions(self):
        return {
            "laser": (":class:`.Laser`", "The laser to set mode power for."),
            **super().argument_descriptions(),
        }


class PhaseConfigDumper(SignatureArgumentMixin, ItemDumper):
    def __init__(self):
        super().__init__(item_type=Model.phase_config)

    def __call__(self, adapter, model):
        # Grab the values from the model private settings (Cython) object but grab the
        # defaults from the :meth:`.Model.phase_config` signature.
        settings = {
            "zero_k00": model._settings.phase_config.zero_k00,
            "zero_tem00_gouy": model._settings.phase_config.zero_tem00_gouy,
        }
        sigparams = self.arguments()
        parameters = {
            key: ArgumentDump(
                key,
                value=value,
                default=sigparams[key].default,
                kind=sigparams[key].kind,
            )
            for key, value in settings.items()
        }

        yield CommandDump(
            adapter=adapter,
            parameters=parameters,
            is_default=all(param.is_default for param in parameters.values()),
        )


class KatSpec:
    """KatScript language specification.

    This defines the available KatScript elements, commands, and analyses that the
    parser recognises. These directives are mapped by this object to adapter classes to
    convert between KatScript and Finesse objects.

    Additional directives can be registered at runtime using :meth:`.register_element`,
    :meth:`.register_command`, and :meth:`.register_analysis`. These adapters will
    remain for the rest of the lifetime of the object.

    Supported KatScript keywords, constants, operators, and functions are not modifiable
    by users.

    This class should not normally be instantiated by user code; the shared instance in
    :data:`.KATSPEC` should instead be used.
    """

    # Fixed language constructs. These are not modifiable by the user.
    keywords = {
        # None.
        "none",
        # HOM collections.
        "even",
        "odd",
        "x",
        "y",
        "off",
        # Axis scales.
        "lin",
        "log",
        # Modulator types.
        "am",
        "pm",
        # Filter types.
        "lowpass",
        "highpass",
        "bandpass",
        "bandstop",
        "single",
        "xsplit",
        "ysplit",
        # Beam properties (see :class:`.BeamProperty`).
        *detectors.bpdetector.BP_KEYWORDS.keys(),
        # Cavity properties (see :class:`.CavityProperty`).
        *detectors.cavity_detector.CP_KEYWORDS.keys(),
        "both",
    }
    constants = symbols.CONSTANTS
    unary_operators = {
        "+": symbols.FUNCTIONS["pos"],
        "-": symbols.FUNCTIONS["neg"],
    }
    binary_operators = {
        "+": symbols.OPERATORS["__add__"],
        "-": symbols.OPERATORS["__sub__"],
        "*": symbols.OPERATORS["__mul__"],
        "**": symbols.OPERATORS["__pow__"],
        "/": symbols.OPERATORS["__truediv__"],
        "//": symbols.OPERATORS["__floordiv__"],
    }
    expression_functions = symbols.FUNCTIONS

    def __init__(self):
        # Modifiable specifications. These are dynamically supported by the parser.
        self.elements = {}
        self.commands = {}
        self.analyses = {}
        self._register_constructs()

    def _register_constructs(self):
        ### Register default language constructs.

        ## Components.
        self.register_element(make_element(electronics.Amplifier, "amplifier"))
        self.register_element(
            make_element(components.Beamsplitter, "beamsplitter", short_name="bs")
        )
        # The cavity factory's `last` flag is set because cavities implicitly depend on
        # any nodes that appear in paths from their start ports back to themselves, so
        # their dependencies cannot be determined by the time the first set of elements
        # are built into the model. It is moved to the second build pass by this flag.
        self.register_element(
            make_element(components.Cavity, "cavity", short_name="cav", build_last=True)
        )
        self.register_element(
            make_element(
                components.DegreeOfFreedom, "degree_of_freedom", short_name="dof"
            )
        )
        self.register_element(
            make_element(
                components.DirectionalBeamsplitter,
                "directional_beamsplitter",
                short_name="dbs",
            )
        )
        self.register_element(
            make_element(electronics.TestPoint, "test_point", short_name="test")
        )
        self.register_element(
            make_element(electronics.ZPKFilter, "filter_zpk", short_name="zpk")
        )
        self.register_element(
            make_element(electronics.ButterFilter, "filter_butter", short_name="butter")
        )
        self.register_element(
            make_element(electronics.Cheby1Filter, "filter_cheby1", short_name="cheby1")
        )
        self.register_element(
            make_element(components.Isolator, "isolator", short_name="isol")
        )
        self.register_element(make_element(components.Laser, "laser", short_name="l"))
        self.register_element(make_element(components.Lens, "lens"))
        self.register_element(make_element(components.AstigmaticLens, "alens"))
        self.register_element(make_element(components.Mirror, "mirror", short_name="m"))
        self.register_element(make_element(components.FrequencyLoss, "floss"))
        self.register_element(
            make_element(components.Modulator, "modulator", short_name="mod")
        )
        self.register_element(
            make_element(
                components.optical_bandpass.OpticalBandpassFilter,
                "optical_bandpass",
                short_name="obp",
            )
        )
        self.register_element(
            make_element(components.Squeezer, "squeezer", short_name="sq")
        )
        self.register_element(make_element(components.ReadoutDC, "readout_dc"))
        self.register_element(make_element(components.ReadoutRF, "readout_rf"))
        self.register_element(
            make_element(
                components.SignalGenerator, "signal_generator", short_name="sgen"
            )
        )
        self.register_element(
            make_element(components.Telescope, "telescope", short_name="tel")
        )

        ## Detectors.
        self.register_element(
            make_element(
                detectors.AmplitudeDetector, "amplitude_detector", short_name="ad"
            )
        )
        self.register_element(
            make_element(detectors.FieldDetector, "field_detector", short_name="fd")
        )
        self.register_element(make_element(detectors.AstigmatismDetector, "astigd"))
        self.register_element(
            make_element(
                detectors.BeamPropertyDetector,
                "beam_property_detector",
                short_name="bp",
            )
        )
        self.register_element(
            make_element(detectors.OptimalQ, "optimal_q_detector", short_name="optbp")
        )
        self.register_element(make_element(detectors.CCD, "ccd"))
        self.register_element(make_element(detectors.CCDScanLine, "ccdline"))
        self.register_element(make_element(detectors.CCDPixel, "ccdpx"))
        self.register_element(
            make_element(
                detectors.CavityPropertyDetector,
                "cavity_property_detector",
                short_name="cp",
            )
        )
        self.register_element(make_element(detectors.FieldCamera, "fcam"))
        self.register_element(make_element(detectors.FieldScanLine, "fline"))
        self.register_element(make_element(detectors.FieldPixel, "fpx"))
        # The gouy detector factory's `last` flag is set it implicitly depends on any
        # nodes that appear in the path from its start port back to itself, so its
        # dependencies cannot be determined by the time the first set of elements are
        # built into the model. It is moved to the second build pass by this flag.
        self.register_element(make_element(detectors.Gouy, "gouy", build_last=True))
        self.register_element(make_element(detectors.KnmDetector, "knmd"))
        self.register_element(make_element(detectors.ModeMismatchDetector, "mmd"))
        self.register_element(
            make_element(detectors.MathDetector, "math_detector", short_name="mathd")
        )
        self.register_element(
            make_element(detectors.MotionDetector, "motion_detector", short_name="xd")
        )
        self.register_element(
            make_element(detectors.PowerDetector, "power_detector_dc", short_name="pd")
        )
        self.register_element(
            make_element(
                detectors.PowerDetectorDemod1,
                "power_detector_demod_1",
                short_name="pd1",
            )
        )
        self.register_element(
            make_element(
                detectors.PowerDetectorDemod2,
                "power_detector_demod_2",
                short_name="pd2",
            )
        )
        self.register_element(
            make_element(
                detectors.QuantumNoiseDetector,
                "quantum_noise_detector",
                short_name="qnoised",
            )
        )
        self.register_element(
            make_element(
                detectors.QuantumNoiseDetectorDemod1,
                "quantum_noise_detector_demod_1",
                short_name="qnoised1",
            )
        )
        self.register_element(
            make_element(
                detectors.QuantumNoiseDetectorDemod2,
                "quantum_noise_detector_demod_2",
                short_name="qnoised2",
            )
        )
        self.register_element(
            make_element(
                detectors.QuantumShotNoiseDetector,
                "quantum_shot_noise_detector",
                short_name="qshot",
            )
        )
        self.register_element(
            make_element(
                detectors.QuantumShotNoiseDetectorDemod1,
                "quantum_shot_noise_detector_demod_1",
                short_name="qshot1",
            )
        )
        self.register_element(
            make_element(
                detectors.QuantumShotNoiseDetectorDemod2,
                "quantum_shot_noise_detector_demod_2",
                short_name="qshot2",
            )
        )

        ## Connectors.
        self.register_element(make_element(components.Space, "space", short_name="s"))
        self.register_element(make_element(components.Wire, "wire", short_name="w"))
        self.register_element(make_element(components.Nothing, "nothing"))

        ## Mechanics.
        self.register_element(make_element(mechanical.FreeMass, "free_mass"))
        self.register_element(make_element(mechanical.Pendulum, "pendulum"))
        self.register_element(
            make_element(
                mechanical.SuspensionZPK, "suspension_zpk", short_name="sus_zpk"
            )
        )

        ## Lock.
        self.register_element(make_element(locks.Lock, "lock", ref_args=("feedback",)))

        ## Noise.
        self.register_element(make_element(noise.ClassicalNoise, "noise"))

        ## Gauss.
        self.register_element(
            ItemAdapter(
                full_name="gauss",
                getter=GaussDumper(),
                factory=ElementFactory(item_type=components.Gauss),
                setter=ElementSetter(item_type=components.Gauss),
                documenter=ElementDocumenter(item_type=components.Gauss),
            )
        )

        ## Variable.
        self.register_element(
            ItemAdapter(
                full_name="variable",
                getter=VariableDumper(),  # ElementDumper(item_type=components.Variable),
                factory=ElementFactory(item_type=components.Variable),
                setter=ElementSetter(item_type=components.Variable),
                documenter=ElementDocumenter(item_type=components.Variable),
                short_name="var",
            )
        )

        ## Commands
        self.register_command(
            ItemAdapter(
                full_name="fsig",
                getter=FsigDumper(),
                setter=FsigSetter(),
                documenter=FsigDocumenter(),
                singular=True,
            )
        )
        self.register_command(
            ItemAdapter(
                full_name="lambda",
                getter=CommandPropertyDumper(
                    item_type=Model.lambda0,
                    default=config_instance()["constants"].getfloat("lambda0"),
                ),
                setter=CommandPropertySetter(item_type=Model.lambda0),
                documenter=CommandPropertyDocumenter(item_type=Model.lambda0),
                singular=True,
            )
        )
        self.register_command(
            ItemAdapter(
                full_name="modes",
                getter=ModesDumper(),
                setter=CommandMethodSetter(item_type=Model.modes),
                documenter=CommandMethodDocumenter(item_type=Model.modes),
                singular=True,
            )
        )
        # self.register_command(
        #     ItemAdapter(
        #         full_name="add_parameter",
        #         getter=CommandMethodSetter(item_type=Model.add_parameter),
        #         setter=CommandMethodSetter(item_type=Model.add_parameter),
        #         documenter=CommandMethodDocumenter(item_type=Model.add_parameter),
        #     )
        # )
        self.register_command(
            ItemAdapter(
                full_name="link",
                setter=CommandMethodSetter(item_type=Model.link),
                documenter=CommandMethodDocumenter(item_type=Model.link),
            )
        )
        self.register_command(
            ItemAdapter(
                full_name="tem",
                getter=TEMDumper(),
                setter=TEMSetter(),
                documenter=TEMDocumenter(),
            )
        )
        self.register_command(
            ItemAdapter(
                full_name="phase_config",
                getter=PhaseConfigDumper(),
                setter=CommandMethodSetter(item_type=Model.phase_config),
                documenter=CommandMethodDocumenter(item_type=Model.phase_config),
            )
        )

        ## Group actions.
        self.register_analysis(make_analysis(actions.Parallel, "parallel"))
        self.register_analysis(make_analysis(actions.Series, "series"))
        self.register_analysis(make_analysis(actions.For, "for"))
        ## Axes.
        self.register_analysis(make_analysis(actions.Noxaxis, "noxaxis"))
        self.register_analysis(
            make_analysis(actions.Xaxis, "xaxis", ref_args=("parameter",))
        )
        self.register_analysis(
            make_analysis(
                actions.X2axis, "x2axis", ref_args=("parameter1", "parameter2")
            )
        )
        self.register_analysis(
            make_analysis(
                actions.X3axis,
                "x3axis",
                ref_args=("parameter1", "parameter2", "parameter3"),
            )
        )
        self.register_analysis(
            make_analysis(actions.Sweep, "sweep", ref_args=("*args",))
        )
        self.register_analysis(make_analysis(actions.Change, "change"))
        self.register_analysis(make_analysis(actions.UpdateMaps, "update_maps"))
        self.register_analysis(
            make_analysis(
                actions.FrequencyResponse, "frequency_response", short_name="freqresp"
            )
        )
        self.register_analysis(
            make_analysis(
                actions.FrequencyResponse2,
                "frequency_response2",
                short_name="freqresp2",
            )
        )
        self.register_analysis(
            make_analysis(
                actions.FrequencyResponse3,
                "frequency_response3",
                short_name="freqresp3",
            )
        )
        self.register_analysis(
            make_analysis(
                actions.FrequencyResponse4,
                "frequency_response4",
                short_name="freqresp4",
            )
        )
        self.register_analysis(
            make_analysis(actions.OptimiseRFReadoutPhaseDC, "opt_rf_readout_phase")
        )
        self.register_analysis(
            make_analysis(actions.SensingMatrixDC, "sensing_matrix_dc")
        )
        self.register_analysis(make_analysis(actions.SetLockGains, "set_lock_gains"))
        self.register_analysis(
            make_analysis(actions.GetErrorSignals, "get_error_signals")
        )
        ## Model physics.
        self.register_analysis(make_analysis(actions.Eigenmodes, "eigenmodes"))
        self.register_analysis(make_analysis(actions.Operator, "operator"))
        self.register_analysis(make_analysis(actions.ABCD, "abcd"))
        self.register_analysis(make_analysis(actions.BeamTrace, "beam_trace"))
        self.register_analysis(make_analysis(actions.PropagateBeam, "propagate_beam"))
        self.register_analysis(
            make_analysis(actions.PropagateAstigmaticBeam, "propagate_beam_astig")
        )
        self.register_analysis(make_analysis(actions.AntiSqueezing, "antisqueezing"))
        ## Utilities.
        self.register_analysis(make_analysis(actions.Debug, "debug"))
        self.register_analysis(make_analysis(actions.debug.SaveMatrix, "save_matrix"))
        self.register_analysis(make_analysis(actions.Plot, "plot"))
        self.register_analysis(make_analysis(actions.Printer, "print"))
        self.register_analysis(make_analysis(actions.RunLocks, "run_locks"))
        self.register_analysis(
            make_analysis(actions.PseudoLockCavity, "pseudo_lock_cavity")
        )
        self.register_analysis(
            make_analysis(actions.PseudoLockDRFPMI, "pseudo_lock_drfpmi")
        )
        self.register_analysis(
            make_analysis(actions.NoiseProjection, "noise_projection")
        )
        self.register_analysis(make_analysis(actions.PrintModel, "print_model"))
        self.register_analysis(
            make_analysis(actions.PrintModelAttr, "print_model_attr")
        )
        self.register_analysis(make_analysis(actions.Minimize, "minimize"))
        self.register_analysis(make_analysis(actions.Maximize, "maximize"))
        self.register_analysis(make_analysis(actions.DCFields, "dc_fields"))

    @property
    def directives(self):
        """All top level parser directives.

        :`getter`: Returns a mapping of top level parser directive aliases to
                   :class:`adapters <.ItemAdapter>`.
        :`type`: :class:`~collections.ChainMap`
        """
        # ChainMap yields in LIFO order so key order becomes elements, then commands,
        # then analyses. This order is relied upon by :func:`.syntax`.
        return ChainMap(self.analyses, self.commands, self.elements)

    @property
    def function_directives(self):
        """All top level function-style parser directives.

        :`getter`: Returns a mapping of top level function parser directive aliases to
                   :class:`adapters <.ItemAdapter>`.
        :`type`: :class:`~collections.ChainMap`
        """
        return ChainMap(self.analyses, self.commands)

    @property
    def reserved_names(self):
        """All reserved names.

        This is primarily useful for tests.

        :`getter`: Returns the names reserved in the parser as special production types.
        :`type`: :class:`list`
        """
        return list(self.keywords) + list(self.constants)

    def _register_adapter(self, mapping, adapter, overwrite=False):
        for alias in adapter.aliases:
            if alias in mapping:
                if overwrite:
                    LOGGER.info(
                        f"overwriting existing {repr(alias)} with {repr(adapter)}"
                    )
                else:
                    raise KeyError(
                        f"{repr(alias)} from {repr(adapter)} already exists (provided "
                        f"by {mapping[alias]}). If you intend to overwrite the "
                        f"existing definition, set overwrite=True."
                    )

            mapping[alias] = adapter

    def register_element(self, adapter, **kwargs):
        """Add parser and generator support for a model element such as a component or
        detector.

        Parameters
        ----------
        adapter : :class:`.ItemAdapter`
            The element adapter.

        Other Parameters
        ----------------
        overwrite : bool, optional
            Overwrite elements with the same aliases, if present. If `False` and one of
            `adapter`'s aliases already exists, a :class:`KeyError` is raised.
            Defaults to `False`.
        """
        self._register_adapter(self.elements, adapter, **kwargs)

    def register_command(self, adapter, **kwargs):
        """Add parser and generator support for a command.

        Parameters
        ----------
        adapter : :class:`.ItemAdapter`
            The command adapter.

        Other Parameters
        ----------------
        overwrite : bool, optional
            Overwrite commands with the same aliases, if present. If `False` and one of
            `adapter`'s aliases already exists, a :class:`KeyError` is raised.
            Defaults to `False`.
        """
        self._register_adapter(self.commands, adapter, **kwargs)

    def register_analysis(self, adapter, **kwargs):
        """Add parser and generator support for an analysis.

        Parameters
        ----------
        adapter : :class:`.ItemAdapter`
            The analysis adapter.

        Other Parameters
        ----------------
        overwrite : bool, optional
            Overwrite analyses with the same aliases, if present. If `False` and one of
            `adapter`'s aliases already exists, a :class:`KeyError` is raised.
            Defaults to `False`.
        """
        self._register_adapter(self.analyses, adapter, **kwargs)

    def type_descriptor(self, _type, default=None):
        """Get a descriptor for a type that's suitable for use in user feedback.

        This allows something other than Python class names to be displayed to the user
        inside error messages.

        Supports the same parameters as :meth:`dict.get`.
        """
        # This function is typically only needed by error handling code, so import
        # required types on first call.
        from ..components.node import Node, Port

        descriptors = {
            str: "string",
            int: "integer",
            float: "floating point",
            complex: "complex",
            Node: "node",
            Port: "port",
        }

        return descriptors.get(_type, default)

    def get_element_class(self, name: str) -> type:
        """Get the corresponding python class for a katscript model element.

        Parameters
        ----------
        name : str
            name of a katscript element

        Returns
        -------
        type
            Python class for this element

        Raises
        ------
        ValueError
            When the element can not be found
        """
        try:
            return self.elements[name].factory.item_type
        except KeyError:
            raise ValueError(f"'{name}' does not represent an element in KatScript")


# Shared KatSpec instance.
KATSPEC = KatSpec()



import pytest

from finesse.utilities.docs import class_to_url, get_docs_url
from finesse.components import Mirror


def test_get_docs_url():
    assert get_docs_url() == "https://finesse.ifosim.org/docs/"


def test_class_to_url():
    import finesse

    if ".dev" in finesse.__version__:
        version = "develop"
    else:
        version = finesse.__version__
    assert (
        class_to_url(Mirror)
        == f"https://finesse.ifosim.org/docs/{version}/api/components/mirror/finesse.components.mirror.Mirror.html"
    )


def test_class_to_url_no_finesse_class():
    with pytest.raises(ValueError):
        class_to_url(int)



#cython: boundscheck=False, wraparound=False, initializedcheck=False

from finesse.cymath cimport complex_t
from finesse.simulations.simulation cimport BaseSimulation
from finesse.cymath.cmatrix cimport SubCCSView1DArray
from finesse.symbols import Symbol
from finesse.frequency cimport frequency_info_t
from finesse.cymath.cmatrix cimport SubCCSView
from finesse.components.workspace cimport FillFuncWrapper
from cpython.ref cimport PyObject

import logging

ctypedef (double*, ) ptr_tuple_1
ctypedef (double*, double*) ptr_tuple_2


LOGGER = logging.getLogger(__name__)


cdef class LensValues(BaseCValues):
    def __init__(LensValues self):
        cdef ptr_tuple_1 ptr = (&self.f, )
        cdef tuple params = ("f", )
        self.setup(params, sizeof(ptr), <double**>&ptr)


cdef class AstigmaticLensValues(BaseCValues):
    def __init__(AstigmaticLensValues self):
        cdef ptr_tuple_2 ptr = (&self.fx, &self.fy)
        cdef tuple params = ("fx", "fy")
        self.setup(params, sizeof(ptr), <double**>&ptr)


cdef class LensConnections:
    def __cinit__(self, BaseSimulation sim):
        self.P1i_P2o = SubCCSView1DArray(sim.carrier.optical_frequencies.size)
        self.P2i_P1o = SubCCSView1DArray(sim.carrier.optical_frequencies.size)

        self.conn_ptrs.P1i_P2o = <PyObject**>self.P1i_P2o.views
        self.conn_ptrs.P2i_P1o = <PyObject**>self.P2i_P1o.views


cdef class BaseLensWorkspace(KnmConnectorWorkspace):
    def __init__(self, owner, BaseSimulation sim, BaseCValues values):
        super().__init__(
            owner,
            sim,
            LensConnections(sim),
            None,
            values,
        )
        self.lc = self.carrier.connections

        self.P1i_id = sim.trace_node_index[owner.p1.i]
        self.P1o_id = sim.trace_node_index[owner.p1.o]
        self.P2i_id = sim.trace_node_index[owner.p2.i]
        self.P2o_id = sim.trace_node_index[owner.p2.o]

        self.sym_abcd_Cx = NULL
        self.sym_abcd_Cy = NULL

    def __dealloc__(self):
        cy_expr_free(self.sym_abcd_Cx)
        cy_expr_free(self.sym_abcd_Cy)

    def compile_abcd_cy_exprs(self):
        cdef:
            dict abcd_handles = self.owner._abcd_matrices
            tuple keyx = (self.owner.p1.i, self.owner.p2.o, "x")
            tuple keyy = (self.owner.p1.i, self.owner.p2.o, "y")
            cdef object[:, ::1] Mx_sym = abcd_handles[keyx][0]
            cdef object[:, ::1] My_sym = abcd_handles[keyy][0]

        # NOTE (sjr) Only element C of a Lens ABCD matrix can possibly change

        if isinstance(Mx_sym[1][0], Symbol):
            ch_sym = Mx_sym[1][0].expand_symbols().eval(keep_changing_symbols=True)
            if isinstance(ch_sym, Symbol):
                self.sym_abcd_Cx = cy_expr_new()
                cy_expr_init(self.sym_abcd_Cx, ch_sym)

        if isinstance(My_sym[1][0], Symbol):
            ch_sym = My_sym[1][0].expand_symbols().eval(keep_changing_symbols=True)
            if isinstance(ch_sym, Symbol):
                self.sym_abcd_Cy = cy_expr_new()
                cy_expr_init(self.sym_abcd_Cy, ch_sym)

    cpdef update_parameter_values(self) :
        ConnectorWorkspace.update_parameter_values(self)

        if self.sym_abcd_Cx != NULL:
            self.abcd_x[1][0] = cy_expr_eval(self.sym_abcd_Cx)

        if self.sym_abcd_Cy != NULL:
            self.abcd_y[1][0] = cy_expr_eval(self.sym_abcd_Cy)


cdef class LensWorkspace(BaseLensWorkspace):
    def __init__(self, owner, BaseSimulation sim):
        self.lv = LensValues()
        super().__init__(owner, sim, self.lv)


cdef class AstigmaticLensWorkspace(BaseLensWorkspace):
    def __init__(self, owner, BaseSimulation sim):
        self.lv = AstigmaticLensValues()
        super().__init__(owner, sim, self.lv)



# TODO (sjr) make c_lens_fill function?

lens_fill_qnoise = FillFuncWrapper.make_from_ptr(c_lens_fill_qnoise)
cdef object c_lens_fill_qnoise(ConnectorWorkspace cws) :
    r"""
    Fills the quantum noise input matrix elements corresponding to this `lens`.
    """
    cdef:
        BaseLensWorkspace ws = <BaseLensWorkspace> cws
        PyObject ***noises = ws.output_noise.ptrs
        frequency_info_t *freq

        Py_ssize_t i, j
        complex_t factor

    for i in range(ws.sim.signal.optical_frequencies.size):
        freq = &(ws.sim.signal.optical_frequencies.frequency_info[i])
        factor = 0.5 * (1 + freq.f_car[0] / ws.sim.model_settings.f0)

        if ws.sim.is_modal:
            ws.total_losses[:] = 0
            for j in range(ws.sim.signal.nhoms):
                ws.total_losses[j] += ws.oconn_info[0].loss[j]
            (<SubCCSView>noises[0][freq.index]).fill_za_dv(factor, ws.total_losses)

            ws.total_losses[:] = 0
            for j in range(ws.sim.signal.nhoms):
                ws.total_losses[j] += ws.oconn_info[1].loss[j]
            (<SubCCSView>noises[1][freq.index]).fill_za_dv(factor, ws.total_losses)



"""Tree data structure."""

from functools import reduce
import networkx as nx

cdef class TreeNode:
    """Container for tree-like structures.

    Tree nodes can contain other tree nodes, allowing these objects to be built into a
    tree-like hierarchy useful for representing connections between components, nested
    data sets, etc.

    Parameters
    ----------
    name : str
        The node name.

    parent: :class:`.TreeNode`, optional
        The parent tree node, if not the root.

    empty : bool, optional
        Whether this node is considered "empty", which determines which character to use
        to represent the node in the tree; defaults to True.

    edge_info : str, optional
        String information on how this node is connected to its parent node. Will be used
        in :meth:`TreeNode.draw_tree`
    """
    def __init__(self, str name, TreeNode parent=None, bint empty=True, str edge_info=None):
        self.__name = name
        self.children = []
        self.empty = empty
        self.edge_info = edge_info
        if parent:
            parent.add(self)
        self.parent = parent

    @classmethod
    def from_network(cls, network, root):
        """Create a tree node hierarchy from an acyclic network.

        Notes
        -----
        Cyclic networks are handled by
        :func:`networkx.algorithms.traversal.depth_first_search.dfs_tree`, so this method does not
        need to detect and avoid such cycles.

        Parameters
        ----------
        network : :class:`networkx.Graph`
            The network that is to be represented as a tree.

        root : hashable
            The network node to use as the root of the tree.

        Raises
        ------
        ValueError
            When the specified network is not a forest.
        """
        # Use networkx to convert the network into tree form.
        tree = nx.dfs_tree(network, root)

        if not nx.is_forest(tree):
            raise ValueError("Network must be a forest (no undirected cycles).")

        def add_children(tree_node, node_name):
            for child in tree.successors(node_name):
                # Verbose component networks store edge information here
                edge_info = network.nodes[child].get(node_name, None)
                child_tree_node = cls(child, parent=tree_node, edge_info=edge_info)
                add_children(child_tree_node, child)

        head_node = cls(root)
        add_children(head_node, root)

        return head_node

    def add(self, child):
        if child is self:
            raise Exception("Can't add self as a node")

        if child in self.children:
            raise Exception(
                f"Child {repr(child)} has already been added to {repr(self)}"
            )
        self.children.append(child)
        child.parent = self

    def get_all_children(self):
        rtn = [*self.children]
        for child in self.children:
            rtn.extend(child.get_all_children())
        return rtn

    def get_all_parents(self):
        rtn = []
        curr = self

        while curr.parent is not None:
            rtn.append(curr.parent)
            curr = curr.parent

        return rtn

    @property
    def name(self):
        return self.__name

    def get_path(self):
        return reduce(
            lambda a, b: (b.name if b.name is not None else "") + "/" + a,
            self.get_all_parents(),
            self.name,
        )

    def get(self, ipath):
        els = ipath.strip("/").split("/")
        curr = self

        for i, el in enumerate(els):
            names = tuple(child.name for child in curr.children)
            try:
                idx = names.index(el)
            except ValueError:
                idx = -1

            if idx > -1:
                curr = curr.children[idx]
            else:
                path = "/".join(els[:(i+1)])
                raise Exception(f"Could not find '{path}'")

        return curr

    def ls(self, ipath="/"):
        me = self.get_path()
        if me is None:
            me = ""
        me = me.strip("/")

        if ipath == "/":
            print(list(c.get_path()[(len(me) + 1) :] for c in self.children))
        else:
            print(list(c.get_path()[(len(me) + 1) :] for c in self.get(ipath).children))

    def draw_tree(self, fn_name=None, title=None, show_max_children=None):
        """Draws the solution tree in text form.

        Based on https://stackoverflow.com/a/49638425/2251982.

        Parameters
        ----------
        fn_name : callable, optional
            Function to return the name of a :class:`.TreeNode` given the :class:`.TreeNode` itself.
            Defaults to using :attr:`.TreeNode.name`.

        title : str, optional
            The tree title. If not specified, no title is printed.

        show_max_children : int, optional
            Maximum number of children to show in the tree; defaults to showing all children.

        Returns
        -------
        str
            The tree in textual form.
        """
        branch = "├─"
        pipe = "│"
        end = "╰─"
        dash = "─"

        def format_name(t: TreeNode) -> str:
            if t.edge_info is None:
                return t.name
            else:
                return f"{t.name} ({t.edge_info})"

        fn_name = fn_name or format_name

        if show_max_children is not None:
            show_max_children = int(show_max_children)

            if show_max_children < 0:
                raise ValueError("show_max_children must be >= 0 or None")

        first = fn_name(self) if self.name is not None else None
        char_empty = lambda x: "○" if x.empty else "●"

        lines = [char_empty(self) + " " + (first or "/")]

        def fill_tree(this, lpad):
            maxed_out = False
            for i, child in enumerate(this.children):
                if (
                    show_max_children
                    and i >= show_max_children
                    and i < (len(this.children) - show_max_children)
                ):
                    if not maxed_out:
                        lines.append(lpad + "·")
                        lines.append(lpad + "·")
                        maxed_out = True
                    continue

                if i == len(this.children) - 1:
                    s = end + dash
                    pad = "   "
                else:
                    s = branch + dash
                    pad = pipe + "  "

                lines.append(str(lpad) + str(s) + str(char_empty(child)) + " " + str(fn_name(child)))
                fill_tree(child, str(lpad) + str(pad))

        fill_tree(self, "")

        treestr = ""
        if title is not None:
            treestr += f"- {title}\n"
        treestr += "\n".join(lines)

        return treestr

    def __str__(self):
        return self.draw_tree()



import pytest

from finesse.components import Mirror
from finesse import Model
from finesse.exceptions import FinesseException


@pytest.fixture
def mirror_1():
    return Mirror("m1")


@pytest.fixture
def mirror_2():
    return Mirror("m2")


def test_add(model, mirror_1):
    mirror_1 = model.add(mirror_1)
    assert hasattr(model, "m1")
    assert "m1" in model.elements
    assert mirror_1 == model.m1


def test_add_multiple_models(mirror_1):
    model_1 = Model()
    model_2 = Model()
    mirror_1 = model_1.add(mirror_1)
    with pytest.raises(FinesseException):
        model_2.add(mirror_1)


def test_add_duplicate(model, mirror_1):
    model.add(mirror_1)
    with pytest.raises(FinesseException):
        model.add(mirror_1)


def test_add_unremovable(model, mirror_1):
    model.add(mirror_1, unremovable=True)
    with pytest.raises(FinesseException):
        model.remove(mirror_1)


def test_add_iterable(model, mirror_1, mirror_2):
    m1, m2 = model.add([mirror_1, mirror_2])
    assert m1 == mirror_1
    assert m2 == mirror_2
    assert mirror_1.name in model.elements
    assert mirror_2.name in model.elements



import pytest
from finesse.symbols import FUNCTIONS
from finesse.script.adapter import (
    ItemDumper,
    SignatureAttributeParameterMixin,
    CommandDump,
)


@pytest.fixture
def fake_command_datastore():
    """Datastore for storing global state in command tests.

    This substitutes the model object in some tests.
    """

    class struct:
        def __init__(self):
            self.a = None
            self.b = None
            self.c = None

    return struct()


@pytest.fixture
def fake_command_datastore_dumper(fake_command_datastore, fake_command_func):
    class FakeCommandDatastoreDumper(SignatureAttributeParameterMixin, ItemDumper):
        def __init__(self):
            super().__init__(item_type=fake_command_func, sig_ignore=("model",))

        def __call__(self, adapter):
            parameters = self.dump_parameters(adapter, fake_command_datastore)

            yield CommandDump(
                adapter=adapter,
                parameters=parameters,
                is_default=all(param.is_default for param in parameters.values()),
            )

    return FakeCommandDatastoreDumper


# Override default command to take two optional arguments.
@pytest.fixture
def fake_command_func(fake_command_datastore):
    def fake_command(model, a, /, b=None, *, c=None):
        fake_command_datastore.a = a
        fake_command_datastore.b = b
        fake_command_datastore.c = c

    return fake_command


# Override default element to take one argument.
@pytest.fixture
def fake_element_cls(fake_element_cls):
    class FakeElement(fake_element_cls):
        def __init__(self, name, a):
            super().__init__(name)
            self.a = a

    return FakeElement


@pytest.fixture
def spec(
    spec,
    set_spec_constructs,
    fake_command_adapter_factory,
    fake_command_datastore_dumper,
    fake_command_func,
    fake_element_adapter_factory,
    fake_element_cls,
):
    spec.register_command(
        fake_command_adapter_factory(
            fake_command_func,
            short_name="fake_cmd",
            getter=fake_command_datastore_dumper(),
        )
    )
    spec.register_element(fake_element_adapter_factory(fake_element_cls))
    # These use Finesse functions directly because they are also used like this in a
    # parametrization below.
    set_spec_constructs(
        "unary_operators",
        {"-": FUNCTIONS["neg"]},
        "expression_functions",
        {"sin": FUNCTIONS["sin"], "arctan2": FUNCTIONS["arctan2"]},
    )

    return spec


@pytest.mark.parametrize(
    "args,kwargs,argument_defaults,prefer_keywords,expected",
    (
        # No optional arguments.
        ([None], {}, False, False, "none"),
        ([None], {}, False, True, "none"),
        ([None], {}, True, False, "none, none, c=none"),
        ([None], {}, True, True, "none, b=none, c=none"),
        ([1], {}, False, False, "1"),
        ([1], {}, False, True, "1"),
        ([1], {}, True, False, "1, none, c=none"),
        ([1], {}, True, True, "1, b=none, c=none"),
        # Optional-as-positional arguments.
        ([1, 2], {}, False, False, "1, 2"),
        ([1, 2], {}, False, True, "1, b=2"),
        ([1, 2], {}, True, False, "1, 2, c=none"),
        ([1, 2], {}, True, True, "1, b=2, c=none"),
        ([1, 2], {"c": 3}, False, False, "1, 2, c=3"),
        ([1, 2], {"c": 3}, False, True, "1, b=2, c=3"),
        ([1, 2], {"c": 3}, True, False, "1, 2, c=3"),
        ([1, 2], {"c": 3}, True, True, "1, b=2, c=3"),
        # Optional-as-keyword arguments.
        ([1], {"b": 2}, False, False, "1, 2"),
        ([1], {"b": 2}, False, True, "1, b=2"),
        ([1], {"b": 2}, True, False, "1, 2, c=none"),
        ([1], {"b": 2}, True, True, "1, b=2, c=none"),
        ([1], {"b": 2, "c": 3}, False, False, "1, 2, c=3"),
        ([1], {"b": 2, "c": 3}, False, True, "1, b=2, c=3"),
        ([1], {"b": 2, "c": 3}, True, False, "1, 2, c=3"),
        ([1], {"b": 2, "c": 3}, True, True, "1, b=2, c=3"),
    ),
)
def test_function(
    unbuilder,
    spec,
    args,
    kwargs,
    argument_defaults,
    prefer_keywords,
    expected,
):
    adapter = spec.commands["fake_command"]
    adapter.setter(None, (args, kwargs))
    script = unbuilder.unbuild(
        next(iter(adapter.getter(adapter))),
        argument_defaults=argument_defaults,
        prefer_keywords=prefer_keywords,
    )
    assert script == f"fake_command({expected})"


@pytest.mark.parametrize(
    "function,prefer_keywords,expected",
    (
        (FUNCTIONS["neg"](1), False, "fake_element myel -1"),
        (FUNCTIONS["neg"](1), True, "fake_element myel a=-1"),
        (FUNCTIONS["sin"](1), False, "fake_element myel sin(1)"),
        (FUNCTIONS["sin"](1), True, "fake_element myel a=sin(1)"),
        (FUNCTIONS["arctan2"](1, 2), False, "fake_element myel arctan2(1, 2)"),
        (FUNCTIONS["arctan2"](1, 2), True, "fake_element myel a=arctan2(1, 2)"),
    ),
)
def test_expression_function(
    unbuilder,
    model,
    element_dump,
    fake_element_cls,
    function,
    prefer_keywords,
    expected,
):
    model.add(fake_element_cls("myel", function))
    dump = next(iter(element_dump("fake_element", fake_element_cls, model)))
    script = unbuilder.unbuild(dump, prefer_keywords=prefer_keywords)
    assert script == expected



import numpy as np

from finesse.components.general import Connector
from finesse.components.workspace import ConnectorWorkspace
from finesse.components.node import NodeDirection, NodeType
from finesse.parameter import float_parameter


class TestPoint(Connector):
    """A simple component which has an arbitrary number of test nodes that can be
    connected to and from.

    Examples
    --------
    You could make an electronic element that has three ports:

    >>> from finesse.components.electronics import TestPoint
    >>> model.add(TestPoint('test', 'A', 'B', 'C'))

    The element is called `test`. This has three ports called A, B, and C, each
    with a single node called `io`, as it can be outputed to inputted to.
    """

    def __init__(self, name, *ports: str):
        super().__init__(name)
        for port in ports:
            port = self._add_port(port, NodeType.ELECTRICAL)
            port._add_node("io", NodeDirection.BIDIRECTIONAL)

    def _get_workspace(self, sim):
        return None


class FilterWorkspace(ConnectorWorkspace):
    pass


@float_parameter("gain", "Gain")
# IMPORTANT: renaming this class impacts the katscript spec and should be avoided!
class Amplifier(Connector):
    def __init__(self, name, gain=1):
        super().__init__(name)
        self.gain = gain

        self._add_port("p1", NodeType.ELECTRICAL)
        self.p1._add_node("i", NodeDirection.INPUT)

        self._add_port("p2", NodeType.ELECTRICAL)
        self.p2._add_node("o", NodeDirection.OUTPUT)

        self._register_node_coupling("P1_P2", self.p1.i, self.p2.o)

    def _get_workspace(self, sim):
        if sim.signal:
            if self.p1.i.full_name not in sim.signal.nodes:
                return
            refill = sim.model.fsig.f.is_changing or any(
                p.is_changing for p in self.parameters
            )
            ws = FilterWorkspace(self, sim)
            ws.signal.add_fill_function(self.fill, refill)
            ws.frequencies = sim.signal.signal_frequencies[self.p1.i].frequencies
            return ws
        else:
            return None

    def fill(self, ws):
        if ws.signal.connections.P1_P2_idx > -1:
            for _ in ws.frequencies:
                with ws.sim.signal.component_edge_fill3(
                    ws.owner_id,
                    ws.signal.connections.P1_P2_idx,
                    0,
                    0,
                ) as mat:
                    mat[:] = ws.values.gain

    def eval(self, f):
        return float(self.gain)


@float_parameter("gain", "Gain")
# IMPORTANT: renaming this class impacts the katscript spec and should be avoided!
class Filter(Connector):
    """This is a generic Filter element that encapsulates some of the Scipy signal
    filter tools. The `sys` attribute is the filter object which can be ZPK, BA, or SOS.

    Parameters
    ----------
    name : str
        Name of element in the model
    gain : Parameter
        Overall floating point value gain to apply to the filter.
    """

    def __init__(self, name, gain=1):
        super().__init__(name)
        self.gain = gain

        self._add_port("p1", NodeType.ELECTRICAL)
        self.p1._add_node("i", NodeDirection.INPUT)

        self._add_port("p2", NodeType.ELECTRICAL)
        self.p2._add_node("o", NodeDirection.OUTPUT)

        self._register_node_coupling("P1_P2", self.p1.i, self.p2.o)

    def _get_workspace(self, sim):
        if sim.signal:
            if self.p1.i.full_name not in sim.signal.nodes:
                return
            refill = sim.model.fsig.f.is_changing or any(
                p.is_changing for p in self.parameters
            )
            ws = FilterWorkspace(self, sim)
            ws.signal.add_fill_function(self.fill, refill)
            ws.frequencies = sim.signal.signal_frequencies[self.p1.i].frequencies
            return ws
        else:
            return None

    def fill(self, ws):
        Hz = self.eval(ws.sim.model_settings.fsig)
        if ws.signal.connections.P1_P2_idx > -1:
            for _ in ws.frequencies:
                with ws.sim.signal.component_edge_fill3(
                    ws.owner_id,
                    ws.signal.connections.P1_P2_idx,
                    0,
                    0,
                ) as mat:
                    mat[:] = Hz

    def bode_plot(self, f=None, n=None, return_axes=False):
        """Plots Bode for this filter.

        Parameters
        ----------
        f : optional
            Frequencies to plot for in Hz (Not radians)
        n : int, optional
            number of points to plot

        Returns
        -------
        axis : Matplotlib axis for plot if return_axes=True
        """
        import matplotlib.pyplot as plt
        import scipy
        import scipy.signal

        if f is not None:
            w = 2 * np.pi * f
        else:
            w = None

        # Need to make sure we are converting any symbolics to numerics before
        # handing over to scipy
        sys = (np.array(_, dtype=complex) for _ in self.sys)
        w, mag, phase = scipy.signal.bode(sys, n=n)

        fig, axs = plt.subplots(2, 1, sharex=True)
        axs[0].semilogx(w / 2 / np.pi, mag)
        axs[0].set_ylabel("Amplitude [dB]")

        axs[1].semilogx(w / 2 / np.pi, phase)
        axs[1].set_xlabel("Frequency [Hz]")
        axs[1].set_ylabel("Phase [Deg]")

        fig.suptitle(f"Bode plot for {self.name}")

        if return_axes:
            return axs


@float_parameter("gain", "Gain")
# IMPORTANT: renaming this class impacts the katscript spec and should be avoided!
class ZPKFilter(Filter):
    """A zero-pole-gain filter element that is used for shaping signals in simulations.
    It is a two port element. `p1` is the input port and `p2` is the output port. Each
    one has a single node: `p1.i` and `p2.o`.

    Parameters
    ----------
    name : str
        Name of element in the model
    z : array_like[float | Symbols]
        A 1D-array of zeros. Use `[]` if none are required. By default these are provided
        in units of radians/s, not Hz.
    p : array_like[float | Symbols]
        A 1D-array of poles. Use `[]` if none are required. By default these are provided
        in units of radians/s, not Hz.
    k : [float | Symbol], optional
        Gain factor for the zeros and poles. If `None` then its value is automatically
        set to generate a unity gain at DC.
    fQ : bool, optional
        When True the zeros and poles can be specified in a tuple of
        (frequency, quality factor) for each pole and zero. This automatically
        adds the complex conjugate pair.
    gain : Parameter
        Overall gain for the filter. Differs from `k` as this is a `Parameter` so
        can be easily switched on/off or varied during a simulation.

    Examples
    --------
    Below are a few examples of using a ZPK filter in a simple simulation and
    plotting the output.

    >>> import finesse
    >>> finesse.init_plotting()
    >>> model = finesse.Model()
    >>> model.parse(\"\"\"
    ... # Finesse always expects some optics to be present
    ... # so we make a laser incident on some photodiode
    ... l l1 P=1
    ... readout_dc PD l1.p1.o
    ... # Amplitude modulate a laser
    ... sgen sig l1.amp
    ...
    ... zpk ZPK_unity [] []
    ... link(PD.DC, ZPK_unity)
    ... ad unity ZPK_unity.p2.o f=fsig
    ...
    ... zpk ZPK_1 [] [-10*2*pi]
    ... link(PD.DC, ZPK_1)
    ... ad zpk1 ZPK_1.p2.o f=fsig
    ...
    ... zpk ZPK_2 [-10*2*pi] []
    ... link(PD.DC, ZPK_2)
    ... ad zpk2 ZPK_2.p2.o f=fsig
    ...
    ... # Using symbolics
    ... variable a 20*2*pi
    ... zpk ZPK_symbol [] [-1j*a, 1j*a] -1
    ... link(PD.DC, ZPK_symbol)
    ... ad symbol ZPK_symbol.p2.o f=fsig
    ...
    ... # Using gain parameter instead of k keeps the unity response at DC but
    ... # just flips the sign
    ... zpk ZPK_symbol2 [] [-1j*a, 1j*a] gain=-1
    ... link(PD.DC, ZPK_symbol2)
    ... ad symbol_gain ZPK_symbol2.p2.o f=fsig
    ...
    ... # Symbolics for an RC low pass filter
    ... variable R 100
    ... variable C 10u
    ... zpk ZPK_RC [] [-1/(R*C)]
    ... link(PD.DC, ZPK_RC)
    ... ad RC ZPK_RC.p2.o f=fsig
    ...
    ... fsig(1)
    ... \"\"\")

    >>> sol = model.run("xaxis(fsig, log, 0.1, 10k, 1000)")
    >>> sol.plot(log=True)
    """

    def __init__(self, name, z, p, k=None, *, fQ=False, gain=1):
        super().__init__(name, gain)
        import cmath

        if k is None:
            k = np.prod(np.abs(p)) / np.prod(np.abs(z))

        root = lambda f, Q: -2 * np.pi * f / (2 * Q) + cmath.sqrt(
            (2 * np.pi * f / (2 * Q)) ** 2 - (2 * np.pi * f) ** 2
        )

        if fQ:
            self.z = []
            for f, Q in z:
                r = root(f, Q)
                self.z.append(r)
                self.z.append(r.conjugate())

            self.p = []
            for f, Q in p:
                r = root(f, Q)
                self.p.append(r)
                self.p.append(r.conjugate())
        else:
            self.z = z
            self.p = p
        self.k = k

    @property
    def sys(self):
        """The scipy `sys` object.

        In this case it is a tuple of (zeros, poles, k). This does not convert any
        symbolics used into numerics.
        """
        return (self.z, self.p, self.k * self.gain)

    def eval(self, f):
        """Calculate the value of this filter over some frequencies.

        Parameters
        ----------
        f : array_like
            Frequencies in units of Hz

        Returns
        -------
        H : array_like
            Complex valued filter output
        """
        from ..utilities import zpk_fresp

        return float(self.gain) * zpk_fresp(self.z, self.p, self.k, 2 * np.pi * f)


@float_parameter("gain", "Gain")
# IMPORTANT: renaming this class impacts the katscript spec and should be avoided!
class ButterFilter(ZPKFilter):
    def __init__(self, name, order, btype, frequency, *, gain=1, analog=True):
        super().__init__(name, [], [], [], gain=gain)
        self.__order = order
        self.__btype = btype
        self.__analog = analog
        self.__frequency = frequency
        self.set_zpk()

    def set_zpk(self):
        import scipy.signal as signal

        z, p, k = signal.butter(
            self.order,
            2 * np.pi * np.array(self.frequency),
            btype=self.btype,
            analog=self.analog,
            output="zpk",
        )
        self.z = z
        self.p = p
        self.k = k

    @property
    def frequency(self):
        return self.__frequency

    @frequency.setter
    def frequency(self, value):
        self.__frequency = value
        self.set_zpk()

    @property
    def order(self):
        return self.__order

    @order.setter
    def order(self, value):
        self.__order = value
        self.set_zpk()

    @property
    def btype(self):
        return self.__btype

    @btype.setter
    def btype(self, value):
        self.__btype = value
        self.set_zpk()

    @property
    def analog(self):
        return self.__analog

    @analog.setter
    def analog(self, value):
        self.__analog = value
        self.set_zpk()


@float_parameter("gain", "Gain")
# IMPORTANT: renaming this class impacts the katscript spec and should be avoided!
class Cheby1Filter(ZPKFilter):
    def __init__(self, name, order, rp, btype, frequency, *, gain=1, analog=True):
        import scipy.signal as signal

        zpk = signal.cheby1(
            order,
            rp,
            2 * np.pi * np.array(frequency),
            btype=btype,
            analog=analog,
            output="zpk",
        )
        super().__init__(name, *zpk, gain=gain)



"""Test parser parameter error rewriting."""

import pytest
from finesse.script.exceptions import KatScriptError
from testutils.text import dedent_multiline, escape_full


@pytest.mark.parametrize(
    "script,error",
    (
        # See #335.
        pytest.param(
            dedent_multiline(
                """
                m m1
                pd pd1 1
                """
            ),
            "\nline 2: invalid type for 'pd' argument 'node': expected node, got integer\n"
            "   1: m m1\n"
            "-->2: pd pd1 1\n"
            "             ^\n"
            "Syntax: pd name node pdtype=none",
            id="pd-node-as-number",
        ),
        pytest.param(
            dedent_multiline(
                """
                m m1
                pd pd1 "m1.p1.o"
                """
            ),
            "\nline 2: invalid type for 'pd' argument 'node': expected node, got string\n"
            "   1: m m1\n"
            '-->2: pd pd1 "m1.p1.o"\n'
            "             ^^^^^^^^^\n"
            "Syntax: pd name node pdtype=none",
            id="pd-node-as-string",
        ),
        pytest.param(
            dedent_multiline(
                """
                m m1
                pd pd1 m1
                """
            ),
            "\nline 2: invalid type for 'pd' argument 'node': expected node, got Mirror\n"
            "   1: m m1\n"
            "-->2: pd pd1 m1\n"
            "             ^^\n"
            "Syntax: pd name node pdtype=none",
            id="pd-node-as-port",
        ),
        # See #339.
        pytest.param(
            dedent_multiline(
                """
                m m1
                m m2
                s s1 m1.p1 1 m2.p1
                """
            ),
            "\nline 3: invalid type for 's' argument 'portB': expected port, got integer\n"
            "   2: m m2\n"
            "-->3: s s1 m1.p1 1 m2.p1\n"
            "                 ^\n"
            "Syntax: s name portA portB L=0.0 nr=1.0 user_gouy_x=none user_gouy_y=none",
            id="space-int-as-port",
        ),
    ),
)
def test_type_error(model, script, error):
    """Incorrect parameter types should display an error that offers allowed types."""
    with pytest.raises(KatScriptError, match=escape_full(error)):
        model.parse(script)
    # try:
    # except Exception as e:
    #     assert e.args[0] == error


@pytest.mark.parametrize(
    "script,error",
    (
        pytest.param(
            dedent_multiline(
                """
                m m1
                pd pd1 m1.p1.o pdtype=none a="b"
                """
            ),
            "\nline 2: 'pd' got an unexpected keyword argument 'a'\n"
            "   1: m m1\n"
            '-->2: pd pd1 m1.p1.o pdtype=none a="b"\n'
            "                                 ^\n"
            "Syntax: pd name node pdtype=none",
            id="unexpected-kwarg",
        ),
        pytest.param(
            dedent_multiline(
                """
                m m1
                pd pd1 m1.p1.o 1 2
                """
            ),
            "\nline 2: 'pd' takes 1 positional argument but 3 were given\n"
            "   1: m m1\n"
            "-->2: pd pd1 m1.p1.o 1 2\n"
            "                     ^ ^\n"
            "Syntax: pd name node pdtype=none",
            id="too-many-args",
        ),
    ),
)
def test_call_value_error(model, script, error):
    """Incorrect call errors."""
    with pytest.raises(KatScriptError, match=escape_full(error)):
        model.parse(script)


@pytest.mark.parametrize(
    "script,error",
    (
        # Single value.
        pytest.param(
            dedent_multiline(
                """
                m m1
                m m2
                s s1 m1.p1 m2.mech
                """
            ),
            "\nline 3: invalid value for 's' argument 'portB': must be an optical port for s1.portB\n"
            "   2: m m2\n"
            "-->3: s s1 m1.p1 m2.mech\n"
            "                 ^^^^^^^\n"
            "Syntax: s name portA portB L=0.0 nr=1.0 user_gouy_x=none user_gouy_y=none",
            id="mirror-mech-as-optical-port",
        ),
        pytest.param(
            dedent_multiline(
                """
                m m1
                m m2
                s m1.p1 m2.p1
                """
            ),
            "\nline 3: 's' missing 1 required positional argument: 'portB'\n"
            "   2: m m2\n"
            "-->3: s m1.p1 m2.p1\n"
            "                    ^\n"
            "Syntax: s name portA portB L=0.0 nr=1.0 user_gouy_x=none user_gouy_y=none",
            id="missing-port",
        ),
    ),
)
def test_init_value_error(model, script, error):
    """Errors emitted within an element's __init__."""
    with pytest.raises(KatScriptError, match=escape_full(error)):
        model.parse(script)
